"""Gantry movement wrapper for hardware and simulation based movement."""
from typing import Optional, List, Dict
from typing_extensions import Protocol as TypingProtocol

from opentrons.types import Point, Mount

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.types import Axis as HardwareAxis
from opentrons_shared_data.errors.exceptions import PositionUnknownError

from opentrons.motion_planning import Waypoint

from ..state import StateView
from ..types import MotorAxis, CurrentWell
from ..errors import MustHomeError, InvalidAxisForRobotType


_MOTOR_AXIS_TO_HARDWARE_AXIS: Dict[MotorAxis, HardwareAxis] = {
    MotorAxis.X: HardwareAxis.X,
    MotorAxis.Y: HardwareAxis.Y,
    MotorAxis.LEFT_Z: HardwareAxis.Z,
    MotorAxis.RIGHT_Z: HardwareAxis.A,
    MotorAxis.LEFT_PLUNGER: HardwareAxis.B,
    MotorAxis.RIGHT_PLUNGER: HardwareAxis.C,
    MotorAxis.EXTENSION_Z: HardwareAxis.Z_G,
    MotorAxis.EXTENSION_JAW: HardwareAxis.G,
}

# The height of the bottom of the pipette nozzle at home position without any tips.
# We rely on this being the same for every OT-3 pipette.
#
# We found this number by peeking at the height that OT3Simulator returns for these pipettes:
#   * Flex Single- and 8-Channel P50
#   * Flex Single-, 8-, and 96-channel P1000
#
# That OT3Simulator return value is what Protocol Engine uses for simulation when Protocol Engine
# is configured to not virtualize pipettes, so this number should match it.
VIRTUAL_MAX_OT3_HEIGHT = 248.0


class GantryMover(TypingProtocol):
    """Abstract class for gantry movement handler."""

    async def get_position(
        self,
        pipette_id: str,
        current_well: Optional[CurrentWell] = None,
        fail_on_not_homed: bool = False,
    ) -> Point:
        """Get the current position of the gantry."""
        ...

    def get_max_travel_z(self, pipette_id: str) -> float:
        """Get the maximum allowed z-height for pipette movement."""
        ...

    async def move_to(
        self, pipette_id: str, waypoints: List[Waypoint], speed: Optional[float]
    ) -> Point:
        """Move the hardware gantry to a waypoint."""
        ...

    async def move_relative(
        self,
        pipette_id: str,
        delta: Point,
        speed: Optional[float],
    ) -> Point:
        """Move the hardware gantry in a relative direction by delta."""
        ...

    async def home(self, axes: Optional[List[MotorAxis]]) -> None:
        """Home the gantry."""
        ...

    async def retract_axis(self, axis: MotorAxis) -> None:
        """Retract the specified axis to its home position."""
        ...

    async def prepare_for_mount_movement(self, mount: Mount) -> None:
        """Retract the 'idle' mount if necessary."""
        ...


class HardwareGantryMover(GantryMover):
    """Hardware API based gantry movement handler."""

    def __init__(self, hardware_api: HardwareControlAPI, state_view: StateView) -> None:
        self._hardware_api = hardware_api
        self._state_view = state_view

    async def get_position(
        self,
        pipette_id: str,
        current_well: Optional[CurrentWell] = None,
        fail_on_not_homed: bool = False,
    ) -> Point:
        """Get the current position of the gantry.

        Args:
            pipette_id: Pipette ID to get location data for.
            current_well: Optional parameter for getting pipette location data, effects critical point.
            fail_on_not_homed: Raise PositionUnknownError if gantry position is not known.
        """
        pipette_location = self._state_view.motion.get_pipette_location(
            pipette_id=pipette_id,
            current_location=current_well,
        )
        try:
            return await self._hardware_api.gantry_position(
                mount=pipette_location.mount.to_hw_mount(),
                critical_point=pipette_location.critical_point,
                fail_on_not_homed=fail_on_not_homed,
            )
        except PositionUnknownError as e:
            raise MustHomeError(message=str(e), wrapping=[e])

    def get_max_travel_z(self, pipette_id: str) -> float:
        """Get the maximum allowed z-height for pipette movement.

        Args:
            pipette_id: Pipette ID to get max travel z-height for.
        """
        hw_mount = self._state_view.pipettes.get_mount(pipette_id).to_hw_mount()
        return self._hardware_api.get_instrument_max_height(mount=hw_mount)

    async def move_to(
        self, pipette_id: str, waypoints: List[Waypoint], speed: Optional[float]
    ) -> Point:
        """Move the hardware gantry to a waypoint."""
        assert len(waypoints) > 0, "Must have at least one waypoint"

        hw_mount = self._state_view.pipettes.get_mount(pipette_id).to_hw_mount()

        for waypoint in waypoints:
            await self._hardware_api.move_to(
                mount=hw_mount,
                abs_position=waypoint.position,
                critical_point=waypoint.critical_point,
                speed=speed,
            )

        return waypoints[-1].position

    async def move_relative(
        self,
        pipette_id: str,
        delta: Point,
        speed: Optional[float],
    ) -> Point:
        """Move the hardware gantry in a relative direction by delta.

        Args:
            pipette_id: Not used in hardware implementation.
            delta: Relative X/Y/Z distance to move gantry.
            speed: Optional speed parameter for the move.
        """
        pipette_location = self._state_view.motion.get_pipette_location(
            pipette_id=pipette_id,
        )
        critical_point = pipette_location.critical_point
        hw_mount = pipette_location.mount.to_hw_mount()
        try:
            await self._hardware_api.move_rel(
                mount=hw_mount,
                delta=delta,
                fail_on_not_homed=True,
                speed=speed,
            )
            point = await self._hardware_api.gantry_position(
                mount=hw_mount,
                critical_point=critical_point,
                fail_on_not_homed=True,
            )
        except PositionUnknownError as e:
            raise MustHomeError(message=str(e), wrapping=[e])

        return point

    async def home(self, axes: Optional[List[MotorAxis]]) -> None:
        """Home the gantry."""
        # TODO(mc, 2022-12-01): this is overly complicated
        # https://opentrons.atlassian.net/browse/RET-1287
        if axes is None:
            await self._hardware_api.home()
        elif axes == [MotorAxis.LEFT_PLUNGER]:
            await self._hardware_api.home_plunger(Mount.LEFT)
        elif axes == [MotorAxis.RIGHT_PLUNGER]:
            await self._hardware_api.home_plunger(Mount.RIGHT)
        elif axes == [MotorAxis.LEFT_Z, MotorAxis.LEFT_PLUNGER]:
            await self._hardware_api.home_z(Mount.LEFT)
            await self._hardware_api.home_plunger(Mount.LEFT)
        elif axes == [MotorAxis.RIGHT_Z, MotorAxis.RIGHT_PLUNGER]:
            await self._hardware_api.home_z(Mount.RIGHT)
            await self._hardware_api.home_plunger(Mount.RIGHT)
        else:
            hardware_axes = [_MOTOR_AXIS_TO_HARDWARE_AXIS[a] for a in axes]
            if self._state_view.config.robot_type == "OT-2 Standard" and any(
                axis not in HardwareAxis.ot2_axes() for axis in hardware_axes
            ):
                raise InvalidAxisForRobotType(
                    f"{axes} includes axes that are not valid for OT-2 Standard robot type"
                )
            # Hardware API will raise error if invalid axes are passed for the type of robot
            await self._hardware_api.home(axes=hardware_axes)

    async def retract_axis(self, axis: MotorAxis) -> None:
        """Retract specified axis."""
        hardware_axis = _MOTOR_AXIS_TO_HARDWARE_AXIS[axis]
        if (
            self._state_view.config.robot_type == "OT-2 Standard"
            and hardware_axis not in HardwareAxis.ot2_axes()
        ):
            raise InvalidAxisForRobotType(
                f"{axis} is not valid for OT-2 Standard robot type"
            )
        await self._hardware_api.retract_axis(axis=hardware_axis)

    async def prepare_for_mount_movement(self, mount: Mount) -> None:
        """Retract the 'idle' mount if necessary."""
        await self._hardware_api.prepare_for_mount_movement(mount)


class VirtualGantryMover(GantryMover):
    """State store based gantry movement handler for simulation/analysis."""

    def __init__(self, state_view: StateView) -> None:
        self._state_view = state_view

    async def get_position(
        self,
        pipette_id: str,
        current_well: Optional[CurrentWell] = None,
        fail_on_not_homed: bool = False,
    ) -> Point:
        """Get the current position of the gantry.

        Args:
            pipette_id: Pipette ID to get position for.
            current_well: Not used in virtual implementation.
            fail_on_not_homed: Not used in virtual implementation.
        """
        origin_deck_point = self._state_view.pipettes.get_deck_point(pipette_id)
        if origin_deck_point is not None:
            origin = Point(
                x=origin_deck_point.x, y=origin_deck_point.y, z=origin_deck_point.z
            )
        else:
            origin = Point(x=0, y=0, z=0)
        return origin

    def get_max_travel_z(self, pipette_id: str) -> float:
        """Get the maximum allowed z-height for pipette movement.

        Args:
            pipette_id: Pipette ID to get instrument height and tip length for.
        """
        if self._state_view.config.robot_type == "OT-2 Standard":
            instrument_height = self._state_view.pipettes.get_instrument_max_height_ot2(
                pipette_id
            )
        else:
            instrument_height = VIRTUAL_MAX_OT3_HEIGHT
        tip_length = self._state_view.tips.get_tip_length(pipette_id)
        return instrument_height - tip_length

    async def move_to(
        self, pipette_id: str, waypoints: List[Waypoint], speed: Optional[float]
    ) -> Point:
        """Move the hardware gantry to a waypoint. No-op in virtual implementation."""
        assert len(waypoints) > 0, "Must have at least one waypoint"
        return waypoints[-1].position

    async def move_relative(
        self,
        pipette_id: str,
        delta: Point,
        speed: Optional[float],
    ) -> Point:
        """Move the hardware gantry in a relative direction by delta.

        Args:
            pipette_id: Pipette ID to get position of for virtual move.
            delta: Relative X/Y/Z distance to move gantry.
            speed: Not used in virtual implementation.
        """
        origin = await self.get_position(pipette_id)
        return origin + delta

    async def home(self, axes: Optional[List[MotorAxis]]) -> None:
        """Home the gantry. No-op in virtual implementation."""
        pass

    async def retract_axis(self, axis: MotorAxis) -> None:
        """Retract the specified axis. No-op in virtual implementation."""
        pass

    async def prepare_for_mount_movement(self, mount: Mount) -> None:
        """Retract the 'idle' mount if necessary."""
        pass


def create_gantry_mover(
    state_view: StateView, hardware_api: HardwareControlAPI
) -> GantryMover:
    """Create a gantry mover."""
    return (
        HardwareGantryMover(hardware_api=hardware_api, state_view=state_view)
        if state_view.config.use_virtual_pipettes is False
        else VirtualGantryMover(state_view=state_view)
    )



"""Contains module command validation functions and module errors for heater-shaker."""

# TODO (spp, 2022-03-22): Move these values to heater-shaker module definition.
HEATER_SHAKER_TEMPERATURE_MIN = 37
HEATER_SHAKER_TEMPERATURE_MAX = 95
HEATER_SHAKER_SPEED_MIN = 200
HEATER_SHAKER_SPEED_MAX = 3000


class InvalidTargetTemperatureError(ValueError):
    """An error raised when attempting to set an invalid target temperature."""


class InvalidTargetSpeedError(ValueError):
    """An error raised when attempting to set an invalid target speed."""


def validate_heater_shaker_temperature(celsius: float) -> float:
    """Verify that the target temperature being set is valid for heater-shaker."""
    if HEATER_SHAKER_TEMPERATURE_MIN <= celsius <= HEATER_SHAKER_TEMPERATURE_MAX:
        return celsius
    else:
        raise InvalidTargetTemperatureError(
            f"Cannot set Heater-Shaker to {celsius} °C."
            f" Valid range is {HEATER_SHAKER_TEMPERATURE_MIN}-"
            f"{HEATER_SHAKER_TEMPERATURE_MAX} °C."
        )


def validate_heater_shaker_speed(rpm: int) -> int:
    """Verify that the target speed is valid for heater-shaker"""
    if HEATER_SHAKER_SPEED_MIN <= rpm <= HEATER_SHAKER_SPEED_MAX:
        return rpm
    else:
        raise InvalidTargetSpeedError(
            f"Cannot set Heater-Shaker to shake at {rpm} rpm. Valid speed range is "
            f"{HEATER_SHAKER_SPEED_MIN}-{HEATER_SHAKER_SPEED_MAX} rpm."
        )



"""Forward door events from a `HardwareControlAPI` into a `ProtocolEngine`."""

from __future__ import annotations

from asyncio import get_running_loop, run_coroutine_threadsafe
from typing import Callable, Optional

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.types import (
    HardwareEvent,
    DoorStateNotification,
    DoorState,
    PauseType,
)

from opentrons.protocol_engine.actions import ActionDispatcher, DoorChangeAction

from ..state import StateStore


_UnsubscribeCallback = Callable[[], None]


class DoorWatcher:
    """Forward door events from a `HardwareControlAPI` into a `ProtocolEngine`."""

    def __init__(
        self,
        state_store: StateStore,
        hardware_api: HardwareControlAPI,
        action_dispatcher: ActionDispatcher,
    ) -> None:
        """Initialize the DoorWatcher.

        Args:
            state_store: The StateStore to check if the protocol queue is running.
            hardware_api: The HardwareControlAPI whose events we will listen for.
                Assumed to be running in a separate thread from action_dispatcher.
            action_dispatcher: The ActionDispatcher to dispatch actions into.
                Assumed to be owned by the same event loop that this
                DoorWatcher was constructed in.
        """
        self._state_store = state_store
        self._hardware_api = hardware_api
        self._action_dispatcher = action_dispatcher
        self._loop = get_running_loop()
        self._unsubscribe_callback: Optional[_UnsubscribeCallback] = None

    def start(self) -> None:
        """Subscribe to hardware events and start forwarding them as PE actions."""
        if self._unsubscribe_callback is None:
            self._unsubscribe_callback = self._hardware_api.register_callback(
                self._handle_hardware_door_event
            )

    def stop(self) -> None:
        """Unsubscribe from hardware events.

        Safe to call more than once.

        Warning:
            This method does not wait for pending event dispatches to complete.
            Depending on timing, there may be straggler hardware events
            that the `ActionDispatcher` will see even after this method returns.
        """
        if self._unsubscribe_callback is not None:
            self._unsubscribe_callback()
            self._unsubscribe_callback = None

    def _handle_hardware_door_event(self, event: HardwareEvent) -> None:
        """Handle a door state hardware event, ensuring thread-safety.

        This is used as a callback for HardwareControlAPI.register_callback(),
        and it's run inside the hardware thread.

        This method will only return after the `ProtocolEngine` processes the event.
        If something else is hogging the event loop thread that the `ProtocolEngine`
        is running in, that may take a moment, and the hardware thread may be blocked.

        This method will deadlock if it's ever run from the same thread that
        owns the event loop that this DoorWatcher was constructed in.
        """
        if isinstance(event, DoorStateNotification):
            coroutine = self._handle_hardware_door_event_async(event)
            future = run_coroutine_threadsafe(coroutine, self._loop)
            # Wait for the dispatch to complete before returning,
            # which is important for ordering guarantees.
            future.result()

    async def _handle_hardware_door_event_async(
        self, event: DoorStateNotification
    ) -> None:
        """Handle a door state hardware event.

        This must run in the event loop that owns self._action_dispatcher, for safety.

        Defined as an async function, so we can use this with
        run_coroutine_threadsafe(), which lets us block until
        the dispatch completes.
        """
        # Throw away this event if this instance has already been stop()'d.
        # This can happen if the stop() happens after the HardwareAPI has sent the event,
        # but before this handler has had its turn to run in the asyncio event loop.
        already_stopped = self._unsubscribe_callback is None
        if already_stopped:
            return

        action = DoorChangeAction(door_state=event.new_state)

        if (
            self._state_store.commands.get_is_running()
            and action.door_state == DoorState.OPEN
            and self._state_store.config.block_on_door_open
        ):
            self._hardware_api.pause(PauseType.PAUSE)

        self._action_dispatcher.dispatch(action)



"""Deck configuration resource provider."""
from typing import List, Set, Tuple

from opentrons_shared_data.deck.dev_types import DeckDefinitionV4, CutoutFixture

from opentrons.types import DeckSlotName

from ..types import (
    AddressableArea,
    AreaType,
    PotentialCutoutFixture,
    DeckPoint,
    Dimensions,
    AddressableOffsetVector,
)
from ..errors import (
    CutoutDoesNotExistError,
    FixtureDoesNotExistError,
    AddressableAreaDoesNotExistError,
    FixtureDoesNotProvideAreasError,
)


def get_cutout_position(cutout_id: str, deck_definition: DeckDefinitionV4) -> DeckPoint:
    """Get the base position of a cutout on the deck."""
    for cutout in deck_definition["locations"]["cutouts"]:
        if cutout_id == cutout["id"]:
            position = cutout["position"]
            return DeckPoint(x=position[0], y=position[1], z=position[2])
    else:
        raise CutoutDoesNotExistError(f"Could not find cutout with name {cutout_id}")


def get_cutout_fixture(
    cutout_fixture_id: str, deck_definition: DeckDefinitionV4
) -> CutoutFixture:
    """Gets cutout fixture from deck that matches the cutout fixture ID provided."""
    for cutout_fixture in deck_definition["cutoutFixtures"]:
        if cutout_fixture["id"] == cutout_fixture_id:
            return cutout_fixture
    raise FixtureDoesNotExistError(
        f"Could not find cutout fixture with name {cutout_fixture_id}"
    )


def get_provided_addressable_area_names(
    cutout_fixture_id: str, cutout_id: str, deck_definition: DeckDefinitionV4
) -> List[str]:
    """Gets a list of the addressable areas provided by the cutout fixture on the cutout."""
    cutout_fixture = get_cutout_fixture(cutout_fixture_id, deck_definition)
    try:
        return cutout_fixture["providesAddressableAreas"][cutout_id]
    except KeyError as exception:
        raise FixtureDoesNotProvideAreasError(
            f"Cutout fixture {cutout_fixture['id']} does not provide addressable areas for {cutout_id}"
        ) from exception


def get_addressable_area_display_name(
    addressable_area_name: str, deck_definition: DeckDefinitionV4
) -> str:
    """Get the display name for an addressable area name."""
    for addressable_area in deck_definition["locations"]["addressableAreas"]:
        if addressable_area["id"] == addressable_area_name:
            return addressable_area["displayName"]
    raise AddressableAreaDoesNotExistError(
        f"Could not find addressable area with name {addressable_area_name}"
    )


def get_potential_cutout_fixtures(
    addressable_area_name: str, deck_definition: DeckDefinitionV4
) -> Tuple[str, Set[PotentialCutoutFixture]]:
    """Given an addressable area name, gets the cutout ID associated with it and a set of potential fixtures."""
    potential_fixtures = []
    for cutout_fixture in deck_definition["cutoutFixtures"]:
        for cutout_id, provided_areas in cutout_fixture[
            "providesAddressableAreas"
        ].items():
            if addressable_area_name in provided_areas:
                potential_fixtures.append(
                    PotentialCutoutFixture(
                        cutout_id=cutout_id,
                        cutout_fixture_id=cutout_fixture["id"],
                        provided_addressable_areas=frozenset(provided_areas),
                    )
                )
    # This following logic is making the assumption that every addressable area can only go on one cutout, though
    # it may have multiple cutout fixtures that supply it on that cutout. If this assumption changes, some of the
    # following logic will have to be readjusted
    if not potential_fixtures:
        raise AddressableAreaDoesNotExistError(
            f"{addressable_area_name} is not provided by any cutout fixtures"
            f" in deck definition {deck_definition['otId']}"
        )
    cutout_id = potential_fixtures[0].cutout_id
    assert all(cutout_id == fixture.cutout_id for fixture in potential_fixtures)
    return cutout_id, set(potential_fixtures)


def get_addressable_area_from_name(
    addressable_area_name: str,
    cutout_position: DeckPoint,
    base_slot: DeckSlotName,
    deck_definition: DeckDefinitionV4,
) -> AddressableArea:
    """Given a name and a cutout position, get an addressable area on the deck."""
    for addressable_area in deck_definition["locations"]["addressableAreas"]:
        if addressable_area["id"] == addressable_area_name:
            area_offset = addressable_area["offsetFromCutoutFixture"]
            position = AddressableOffsetVector(
                x=area_offset[0] + cutout_position.x,
                y=area_offset[1] + cutout_position.y,
                z=area_offset[2] + cutout_position.z,
            )
            bounding_box = Dimensions(
                x=addressable_area["boundingBox"]["xDimension"],
                y=addressable_area["boundingBox"]["yDimension"],
                z=addressable_area["boundingBox"]["zDimension"],
            )

            return AddressableArea(
                area_name=addressable_area["id"],
                area_type=AreaType(addressable_area["areaType"]),
                base_slot=base_slot,
                display_name=addressable_area["displayName"],
                bounding_box=bounding_box,
                position=position,
                compatible_module_types=addressable_area.get(
                    "compatibleModuleTypes", []
                ),
            )
    raise AddressableAreaDoesNotExistError(
        f"Could not find addressable area with name {addressable_area_name}"
    )



"""Asynchronous task queue to accomplish a protocol run."""
import asyncio
import logging
from functools import partial
from typing import Any, Awaitable, Callable, Optional
from typing_extensions import Protocol as Callback

log = logging.getLogger(__name__)


class CleanupFunc(Callback):
    """Expected cleanup function signature."""

    def __call__(
        self,
        error: Optional[Exception],
    ) -> Any:
        """Cleanup, optionally taking an error thrown.

        Return value will not be used.
        """
        ...


class TaskQueue:
    """A queue of async tasks to run.

    Once started, a TaskQueue may not be re-used.
    """

    def __init__(
        self,
        # cleanup_func: CleanupFunc,
    ) -> None:
        """Initialize the TaskQueue.

        Args:
            cleanup_func: A function to call at run function completion
                with any error raised by the run function.
        """
        self._cleanup_func: Optional[
            Callable[[Optional[Exception]], Any]
        ] = None  # CleanupFunc = cleanup_func

        self._run_func: Optional[Callable[[], Any]] = None
        self._run_task: Optional["asyncio.Task[None]"] = None
        self._ok_to_join_event: asyncio.Event = asyncio.Event()

    def set_cleanup_func(
        self,
        func: Callable[..., Awaitable[Any]],
        **kwargs: Any,
    ) -> None:
        """Add the protocol cleanup task to the queue.

        The "cleanup" task will be run after the "run" task.
        """
        self._cleanup_func = partial(func, **kwargs)

    def set_run_func(
        self,
        func: Callable[..., Awaitable[Any]],
        **kwargs: Any,
    ) -> None:
        """Add the protocol run task to the queue.

        The "run" task will be run first, before the "cleanup" task.
        """
        self._run_func = partial(func, **kwargs)

    def start(self) -> None:
        """Start running tasks in the queue."""
        self._ok_to_join_event.set()

        if self._run_task is None:
            self._run_task = asyncio.create_task(self._run())

    async def join(self) -> None:
        """Wait for the background run task to complete, propagating errors."""
        await self._ok_to_join_event.wait()

        if self._run_task:
            await self._run_task

    async def _run(self) -> None:
        error = None

        try:
            if self._run_func is not None:
                await self._run_func()
        except Exception as e:
            log.exception("Exception raised by protocol")
            error = e

        if self._cleanup_func is not None:
            await self._cleanup_func(error)



"""Simple state change notification interface."""
import asyncio


class ChangeNotifier:
    """An interface tto emit or subscribe to state change notifications."""

    def __init__(self) -> None:
        """Initialize the ChangeNotifier with an internal Event."""
        self._event = asyncio.Event()

    def notify(self) -> None:
        """Notify all `wait`'ers that the state has changed."""
        self._event.set()

    async def wait(self) -> None:
        """Wait until the next state change notification."""
        self._event.clear()
        await self._event.wait()



"""Waypoint planning."""
from typing import List, Optional, Sequence, Tuple
from typing_extensions import Final

from opentrons.types import Point
from opentrons.hardware_control.types import CriticalPoint

from .types import Waypoint, MoveType, GripperMovementWaypointsWithJawStatus
from .errors import DestinationOutOfBoundsError, ArcOutOfBoundsError
from ..protocol_engine.types import LabwareMovementOffsetData

DEFAULT_GENERAL_ARC_Z_MARGIN: Final[float] = 10.0
DEFAULT_IN_LABWARE_ARC_Z_MARGIN: Final[float] = 5.0
MINIMUM_Z_MARGIN: Final[float] = 1.0


def get_waypoints(
    origin: Point,
    dest: Point,
    *,
    max_travel_z: float,
    min_travel_z: float,
    move_type: MoveType = MoveType.GENERAL_ARC,
    xy_waypoints: Sequence[Tuple[float, float]] = (),
    origin_cp: Optional[CriticalPoint] = None,
    dest_cp: Optional[CriticalPoint] = None,
) -> List[Waypoint]:
    """Get waypoints between an origin point and a destination point.

    Given a move type and Z limits, which should be calculated according to
    deck / labware / pipette geometry, creates waypoints with proper
    z-clearances to move between `origin` and `dest`.

    :param origin: The start point of the move.
    :param dest: The end point of the move.
    :param max_travel_z: The maximum allowed travel height of an arc move.
    :param min_travel_z: The minimum allowed travel height of an arc move.
    :param move_type: Direct move, in-labware arc, or general arc move type.
    :param xy_waypoints: Extra XY destination waypoints to place in the path.
    :param origin_cp: Pipette critical point override for origin waypoints.
    :param dest_cp: Pipette critical point override for destination waypoints.

    :returns: A list of :py:class:`.Waypoint` locations to move through.
    """
    # NOTE(mm, 2022-06-22):
    # This function is used by v6+ JSON protocols and v3+
    # Python API protocols, but not v2 Python API protocols.
    #
    # To experiment with using this module in PAPIv2,
    # flip the default of `use_experimental_waypoint_planning` to True
    # in opentrons.protocols.geometry.planning.plan_moves
    dest_waypoint = Waypoint(dest, dest_cp)
    waypoints: List[Waypoint] = []

    # a direct move can ignore all arc and waypoint planning
    if move_type == MoveType.DIRECT:
        # TODO(mm, 2022-06-17): This will not raise an out-of-bounds error
        # even if the destination is far out of bounds. A protocol can run into this by
        # doing a direct move to bad coordinates. Should we raise in that case?
        return [dest_waypoint]

    # ensure destination is not out of bounds
    if dest.z + MINIMUM_Z_MARGIN > max_travel_z:
        raise DestinationOutOfBoundsError(
            origin=origin,
            dest=dest,
            clearance=MINIMUM_Z_MARGIN,
            min_travel_z=min_travel_z,
            max_travel_z=max_travel_z,
            message="Destination out of bounds in the Z-axis",
        )

    # ensure that the passed in min_travel_z and max_travel_z are compatible
    if min_travel_z + MINIMUM_Z_MARGIN > max_travel_z:
        raise ArcOutOfBoundsError(
            origin=origin,
            dest=dest,
            clearance=MINIMUM_Z_MARGIN,
            min_travel_z=min_travel_z,
            max_travel_z=max_travel_z,
            message="Arc out of bounds in the Z-axis",
        )

    # set the z clearance according to the arc type
    travel_z_margin = (
        DEFAULT_GENERAL_ARC_Z_MARGIN
        if move_type == MoveType.GENERAL_ARC
        else DEFAULT_IN_LABWARE_ARC_Z_MARGIN
    )

    # set the actual travel z according to:
    # use the max of min_travel_z with clearance or dest height
    # if either of those exceed max_travel_z, just use max_travel_z
    # if max_travel_z does not provide enough clearance, check above would
    # raise an ArcOutOfBoundsError
    # if origin.z is higher than the selected travel z, travel at origin.z instead
    travel_z = max(
        min(max_travel_z, max(min_travel_z + travel_z_margin, dest.z)),
        origin.z,
    )

    # if origin.z isn't the travel height: add waypoint to move to origin.z
    if travel_z > origin.z:
        waypoints.append(Waypoint(origin._replace(z=travel_z), origin_cp))

    # add any additional waypoints along with critical point blending
    # see https://github.com/Opentrons/opentrons/pull/5662
    # TODO(mc, 2020-11-05): if any critical point transitions can move in the
    # Z axis, an extra waypoint for that transition will be needed
    for x, y in xy_waypoints:
        waypoints.append(Waypoint(Point(x=x, y=y, z=travel_z), dest_cp))

    # if dest.z isn't the travel height: add waypoint to move to dest.z
    # TODO(mc, 2020-11-05): if any critical point transitions can move in the
    # Z axis, this conditional will need to be revised
    if travel_z > dest.z:
        waypoints.append(Waypoint(dest._replace(z=travel_z), dest_cp))

    waypoints.append(dest_waypoint)

    return waypoints


def get_gripper_labware_movement_waypoints(
    from_labware_center: Point,
    to_labware_center: Point,
    gripper_home_z: float,
    offset_data: LabwareMovementOffsetData,
    post_drop_slide_offset: Optional[Point],
) -> List[GripperMovementWaypointsWithJawStatus]:
    """Get waypoints for moving labware using a gripper."""
    pick_up_offset = offset_data.pickUpOffset
    drop_offset = offset_data.dropOffset

    pick_up_location = from_labware_center + Point(
        pick_up_offset.x, pick_up_offset.y, pick_up_offset.z
    )
    drop_location = to_labware_center + Point(
        drop_offset.x, drop_offset.y, drop_offset.z
    )

    post_drop_home_pos = Point(drop_location.x, drop_location.y, gripper_home_z)

    waypoints_with_jaw_status = [
        GripperMovementWaypointsWithJawStatus(
            position=Point(pick_up_location.x, pick_up_location.y, gripper_home_z),
            jaw_open=False,
            dropping=False,
        ),
        GripperMovementWaypointsWithJawStatus(
            position=pick_up_location, jaw_open=True, dropping=False
        ),
        # Gripper grips the labware here
        GripperMovementWaypointsWithJawStatus(
            position=Point(pick_up_location.x, pick_up_location.y, gripper_home_z),
            jaw_open=False,
            dropping=False,
        ),
        GripperMovementWaypointsWithJawStatus(
            position=Point(drop_location.x, drop_location.y, gripper_home_z),
            jaw_open=False,
            dropping=False,
        ),
        GripperMovementWaypointsWithJawStatus(
            position=drop_location, jaw_open=False, dropping=False
        ),
        # Gripper ungrips here
        GripperMovementWaypointsWithJawStatus(
            position=post_drop_home_pos,
            jaw_open=True,
            dropping=True,
        ),
    ]
    if post_drop_slide_offset is not None:
        # IF it is specified, add one more step after homing the gripper
        waypoints_with_jaw_status.append(
            GripperMovementWaypointsWithJawStatus(
                position=post_drop_home_pos + post_drop_slide_offset,
                jaw_open=True,
                dropping=False,
            )
        )
    return waypoints_with_jaw_status



"""Pipetting command handling."""
from typing import Optional, Iterator
from typing_extensions import Protocol as TypingProtocol
from contextlib import contextmanager

from opentrons.hardware_control import HardwareControlAPI

from ..state import StateView, HardwarePipette
from ..errors.exceptions import (
    TipNotAttachedError,
    InvalidAspirateVolumeError,
    InvalidPushOutVolumeError,
    InvalidDispenseVolumeError,
)


# 1e-9 µL (1 femtoliter!) is a good value because:
# * It's large relative to rounding errors that occur in practice in protocols. For
#   example, https://opentrons.atlassian.net/browse/RESC-182 shows a rounding error
#   on the order of 1e-15 µL.
# * It's small relative to volumes that our users might actually care about and
#   expect the robot to execute faithfully.
# * It's the default absolute tolerance for math.isclose(), where it apparently works
#   well in general.
_VOLUME_ROUNDING_ERROR_TOLERANCE = 1e-9


class PipettingHandler(TypingProtocol):
    """Liquid handling commands."""

    def get_is_ready_to_aspirate(self, pipette_id: str) -> bool:
        """Get whether a pipette is ready to aspirate."""

    async def prepare_for_aspirate(self, pipette_id: str) -> None:
        """Prepare for pipette aspiration."""

    async def aspirate_in_place(
        self,
        pipette_id: str,
        volume: float,
        flow_rate: float,
    ) -> float:
        """Set flow-rate and aspirate."""

    async def dispense_in_place(
        self,
        pipette_id: str,
        volume: float,
        flow_rate: float,
        push_out: Optional[float],
    ) -> float:
        """Set flow-rate and dispense."""

    async def blow_out_in_place(
        self,
        pipette_id: str,
        flow_rate: float,
    ) -> None:
        """Set flow rate and blow-out."""


class HardwarePipettingHandler(PipettingHandler):
    """Liquid handling, using the Hardware API.""" ""

    def __init__(self, state_view: StateView, hardware_api: HardwareControlAPI) -> None:
        """Initialize a PipettingHandler instance."""
        self._state_view = state_view
        self._hardware_api = hardware_api

    def get_is_ready_to_aspirate(self, pipette_id: str) -> bool:
        """Get whether a pipette is ready to aspirate."""
        hw_pipette = self._state_view.pipettes.get_hardware_pipette(
            pipette_id=pipette_id,
            attached_pipettes=self._hardware_api.attached_instruments,
        )
        return (
            self._state_view.pipettes.get_aspirated_volume(pipette_id) is not None
            and hw_pipette.config["ready_to_aspirate"]
        )

    async def prepare_for_aspirate(self, pipette_id: str) -> None:
        """Prepare for pipette aspiration."""
        hw_mount = self._state_view.pipettes.get_mount(pipette_id).to_hw_mount()
        await self._hardware_api.prepare_for_aspirate(mount=hw_mount)

    async def aspirate_in_place(
        self,
        pipette_id: str,
        volume: float,
        flow_rate: float,
    ) -> float:
        """Set flow-rate and aspirate."""
        # get mount and config data from state and hardware controller
        adjusted_volume = _validate_aspirate_volume(
            state_view=self._state_view, pipette_id=pipette_id, aspirate_volume=volume
        )
        hw_pipette = self._state_view.pipettes.get_hardware_pipette(
            pipette_id=pipette_id,
            attached_pipettes=self._hardware_api.attached_instruments,
        )
        with self._set_flow_rate(pipette=hw_pipette, aspirate_flow_rate=flow_rate):
            await self._hardware_api.aspirate(
                mount=hw_pipette.mount, volume=adjusted_volume
            )

        return adjusted_volume

    async def dispense_in_place(
        self,
        pipette_id: str,
        volume: float,
        flow_rate: float,
        push_out: Optional[float],
    ) -> float:
        """Dispense liquid without moving the pipette."""
        adjusted_volume = _validate_dispense_volume(
            state_view=self._state_view, pipette_id=pipette_id, dispense_volume=volume
        )
        hw_pipette = self._state_view.pipettes.get_hardware_pipette(
            pipette_id=pipette_id,
            attached_pipettes=self._hardware_api.attached_instruments,
        )
        # TODO (tz, 8-23-23): add a check for push_out not larger that the max volume allowed when working on this https://opentrons.atlassian.net/browse/RSS-329
        if push_out and push_out < 0:
            raise InvalidPushOutVolumeError(
                "push out value cannot have a negative value."
            )
        with self._set_flow_rate(pipette=hw_pipette, dispense_flow_rate=flow_rate):
            await self._hardware_api.dispense(
                mount=hw_pipette.mount, volume=adjusted_volume, push_out=push_out
            )

        return adjusted_volume

    async def blow_out_in_place(
        self,
        pipette_id: str,
        flow_rate: float,
    ) -> None:
        """Set flow rate and blow-out."""
        # get mount and config data from state and hardware controller
        hw_pipette = self._state_view.pipettes.get_hardware_pipette(
            pipette_id=pipette_id,
            attached_pipettes=self._hardware_api.attached_instruments,
        )
        with self._set_flow_rate(pipette=hw_pipette, blow_out_flow_rate=flow_rate):
            await self._hardware_api.blow_out(mount=hw_pipette.mount)

    @contextmanager
    def _set_flow_rate(
        self,
        pipette: HardwarePipette,
        aspirate_flow_rate: Optional[float] = None,
        dispense_flow_rate: Optional[float] = None,
        blow_out_flow_rate: Optional[float] = None,
    ) -> Iterator[None]:
        """Context manager for setting flow rate before calling aspirate, dispense, or blowout."""
        original_aspirate_rate = pipette.config["aspirate_flow_rate"]
        original_dispense_rate = pipette.config["dispense_flow_rate"]
        original_blow_out_rate = pipette.config["blow_out_flow_rate"]
        self._hardware_api.set_flow_rate(
            pipette.mount,
            aspirate=aspirate_flow_rate,
            dispense=dispense_flow_rate,
            blow_out=blow_out_flow_rate,
        )
        try:
            yield
        finally:
            self._hardware_api.set_flow_rate(
                pipette.mount,
                aspirate=original_aspirate_rate,
                dispense=original_dispense_rate,
                blow_out=original_blow_out_rate,
            )


class VirtualPipettingHandler(PipettingHandler):
    """Liquid handling, using the virtual pipettes.""" ""

    _state_view: StateView

    def __init__(
        self,
        state_view: StateView,
    ) -> None:
        """Initialize a PipettingHandler instance."""
        self._state_view = state_view

    def get_is_ready_to_aspirate(self, pipette_id: str) -> bool:
        """Get whether a pipette is ready to aspirate."""
        return self._state_view.pipettes.get_aspirated_volume(pipette_id) is not None

    async def prepare_for_aspirate(self, pipette_id: str) -> None:
        """Virtually prepare to aspirate (no-op)."""

    async def aspirate_in_place(
        self,
        pipette_id: str,
        volume: float,
        flow_rate: float,
    ) -> float:
        """Virtually aspirate (no-op)."""
        self._validate_tip_attached(pipette_id=pipette_id, command_name="aspirate")
        return _validate_aspirate_volume(
            state_view=self._state_view, pipette_id=pipette_id, aspirate_volume=volume
        )

    async def dispense_in_place(
        self,
        pipette_id: str,
        volume: float,
        flow_rate: float,
        push_out: Optional[float],
    ) -> float:
        """Virtually dispense (no-op)."""
        # TODO (tz, 8-23-23): add a check for push_out not larger that the max volume allowed when working on this https://opentrons.atlassian.net/browse/RSS-329
        if push_out and push_out < 0:
            raise InvalidPushOutVolumeError(
                "push out value cannot have a negative value."
            )
        self._validate_tip_attached(pipette_id=pipette_id, command_name="dispense")
        return _validate_dispense_volume(
            state_view=self._state_view, pipette_id=pipette_id, dispense_volume=volume
        )

    async def blow_out_in_place(
        self,
        pipette_id: str,
        flow_rate: float,
    ) -> None:
        """Virtually blow out (no-op)."""

    def _validate_tip_attached(self, pipette_id: str, command_name: str) -> None:
        """Validate if there is a tip attached."""
        tip_geometry = self._state_view.pipettes.get_attached_tip(pipette_id)
        if not tip_geometry:
            raise TipNotAttachedError(
                f"Cannot perform {command_name} without a tip attached"
            )


def create_pipetting_handler(
    state_view: StateView, hardware_api: HardwareControlAPI
) -> PipettingHandler:
    """Create a pipetting handler."""
    return (
        HardwarePipettingHandler(state_view=state_view, hardware_api=hardware_api)
        if state_view.config.use_virtual_pipettes is False
        else VirtualPipettingHandler(state_view=state_view)
    )


def _validate_aspirate_volume(
    state_view: StateView, pipette_id: str, aspirate_volume: float
) -> float:
    """Get whether the given volume is valid to aspirate right now.

    Return the volume to aspirate, possibly clamped, or raise an
    InvalidAspirateVolumeError.
    """
    working_volume = state_view.pipettes.get_working_volume(pipette_id=pipette_id)

    current_volume = (
        state_view.pipettes.get_aspirated_volume(pipette_id=pipette_id) or 0
    )

    # TODO(mm, 2024-01-11): We should probably just use
    # state_view.pipettes.get_available_volume()? Its whole `None` return vs. exception
    # raising thing is confusing me.
    available_volume = working_volume - current_volume
    available_volume_with_tolerance = (
        available_volume + _VOLUME_ROUNDING_ERROR_TOLERANCE
    )

    if aspirate_volume > available_volume_with_tolerance:
        raise InvalidAspirateVolumeError(
            attempted_aspirate_volume=aspirate_volume,
            available_volume=available_volume,
            max_pipette_volume=state_view.pipettes.get_maximum_volume(
                pipette_id=pipette_id
            ),
            max_tip_volume=_get_max_tip_volume(
                state_view=state_view, pipette_id=pipette_id
            ),
        )
    else:
        return min(aspirate_volume, available_volume)


def _validate_dispense_volume(
    state_view: StateView, pipette_id: str, dispense_volume: float
) -> float:
    """Get whether the given volume is valid to dispense right now.

    Return the volume to dispense, possibly clamped, or raise an
    InvalidDispenseVolumeError.
    """
    aspirated_volume = state_view.pipettes.get_aspirated_volume(pipette_id)
    if aspirated_volume is None:
        raise InvalidDispenseVolumeError(
            "Cannot perform a dispense if there is no volume in attached tip."
        )
    else:
        remaining = aspirated_volume - dispense_volume
        if remaining < -_VOLUME_ROUNDING_ERROR_TOLERANCE:
            raise InvalidDispenseVolumeError(
                f"Cannot dispense {dispense_volume} µL when only {aspirated_volume} µL has been aspirated."
            )
        else:
            return min(dispense_volume, aspirated_volume)


def _get_max_tip_volume(state_view: StateView, pipette_id: str) -> Optional[float]:
    attached_tip = state_view.pipettes.get_attached_tip(pipette_id=pipette_id)
    return None if attached_tip is None else attached_tip.volume



import json
from pydantic import ValidationError
from typing import Optional, Union

from opentrons import config, types
from dataclasses import asdict
from opentrons.util.helpers import utc_now


from .. import file_operators as io, types as local_types

from .models import v1


# Delete Gripper Offset Calibrations


def delete_gripper_calibration_file(gripper: str) -> None:
    """
    Delete gripper calibration offset file based on gripper serial number

    :param gripper: gripper serial number
    """
    offset_path = (
        config.get_opentrons_path("gripper_calibration_dir") / f"{gripper}.json"
    )

    io.delete_file(offset_path)


def clear_gripper_calibration_offsets() -> None:
    """
    Delete all gripper calibration data files.
    """

    offset_dir = config.get_opentrons_path("gripper_calibration_dir")
    io._remove_json_files_in_directories(offset_dir)


# Save Gripper Offset Calibrations


def save_gripper_calibration(
    offset: types.Point,
    gripper_id: str,
    cal_status: Optional[
        Union[local_types.CalibrationStatus, v1.CalibrationStatus]
    ] = None,
) -> None:
    gripper_dir = config.get_opentrons_path("gripper_calibration_dir")

    if isinstance(cal_status, local_types.CalibrationStatus):
        cal_status_model = v1.CalibrationStatus(**asdict(cal_status))
    elif isinstance(cal_status, v1.CalibrationStatus):
        cal_status_model = cal_status
    else:
        cal_status_model = v1.CalibrationStatus()

    gripper_calibration = v1.InstrumentOffsetModel(
        offset=offset,
        lastModified=utc_now(),
        source=local_types.SourceType.user,
        status=cal_status_model,
    )
    io.save_to_file(gripper_dir, gripper_id, gripper_calibration)


# Get Gripper Offset Calibrations


def get_gripper_calibration_offset(
    gripper_id: str,
) -> Optional[v1.InstrumentOffsetModel]:
    """
    Return the requested gripper offset data.
    """
    if not config.feature_flags.enable_ot3_hardware_controller():
        raise NotImplementedError("Gripper calibrations are only valid on the OT-3")
    try:
        gripper_calibration_filepath = (
            config.get_opentrons_path("gripper_calibration_dir") / f"{gripper_id}.json"
        )
        return v1.InstrumentOffsetModel(
            **io.read_cal_file(gripper_calibration_filepath)
        )
    except FileNotFoundError:
        return None
    except (json.JSONDecodeError, ValidationError):
        return None



"""Manager for the :py:class:`.hardware_control.API` thread."""
import threading
import logging
import asyncio
import inspect
import functools
import weakref
from typing import (
    Any,
    Awaitable,
    Callable,
    Generic,
    Optional,
    TypeVar,
    cast,
    Sequence,
    Mapping,
    AsyncGenerator,
    Union,
    Type,
    ParamSpec,
)
from .adapters import SynchronousAdapter
from .modules.mod_abc import AbstractModule
from .protocols import (
    AsyncioConfigurable,
)

MODULE_LOG = logging.getLogger(__name__)


class ThreadManagerException(Exception):
    pass


WrappedReturn = TypeVar("WrappedReturn", contravariant=True)
WrappedYield = TypeVar("WrappedYield", contravariant=True)
P = ParamSpec("P")


async def call_coroutine_threadsafe(
    loop: asyncio.AbstractEventLoop,
    coro: Callable[P, Awaitable[WrappedReturn]],
    *args: P.args,
    **kwargs: P.kwargs,
) -> WrappedReturn:
    fut = cast(
        "asyncio.Future[WrappedReturn]",
        asyncio.run_coroutine_threadsafe(coro(*args, **kwargs), loop),
    )
    wrapped = asyncio.wrap_future(fut)
    return await wrapped


async def execute_asyncgen_threadsafe(
    loop: asyncio.AbstractEventLoop,
    agenfunc: Callable[P, AsyncGenerator[WrappedYield, None]],
    *args: P.args,
    **kwargs: P.kwargs,
) -> AsyncGenerator[WrappedYield, None]:

    # This function should bridge an async generator function between two asyncio
    # loops running in different threads. There are several stages to this because
    # there are several stages to generator execution.

    # These clues will help us later
    class _DoneSingleton:
        pass

    async def _build_queue() -> "asyncio.Queue[Union[WrappedYield, _DoneSingleton]]":
        return asyncio.Queue(maxsize=1)

    yield_queue = await asyncio.wrap_future(
        cast(
            "asyncio.Future[asyncio.Queue[Union[WrappedYield, _DoneSingleton]]]",
            asyncio.run_coroutine_threadsafe(_build_queue(), loop),
        )
    )

    # the async generator function needs to run on the target loop. it is not a coroutine
    # and cannot be run with run_coroutine_threadsafe in the same way. however, we can
    # make a coroutine that _can_, and that exhausts the async generator and puts what
    # the generator yields into a queue. since something later will have to combine
    # awaiting yield results and the function finishing, we can now use the _DoneSingleton
    # to short circuit the caller waiting for a yield result when the function is done.
    # in addition, the queue being 1 element max should give us similar "backpressure"
    # behavior as an async for.
    async def _inner_agen_wrap() -> None:
        item: WrappedYield
        try:
            async for item in agenfunc(*args, **kwargs):
                await yield_queue.put(item)
        finally:
            await yield_queue.put(_DoneSingleton())

    # as promised, we can run our coroutine pretty easily now. note that this coroutine
    # returns None, because it is handling the generator yields internally.
    fut = cast(
        "asyncio.Future[None]",
        asyncio.run_coroutine_threadsafe(_inner_agen_wrap(), loop),
    )

    # now we have to bridge the output of the generator to the calling loop. this is
    # fun because this uses an asyncio queue which isn't thread safe. so we also need
    # to pull stuff out of the asyncio queue via the other loop
    while not fut.done():
        getter = cast(
            "asyncio.Future[Union[WrappedYield, _DoneSingleton]]",
            asyncio.run_coroutine_threadsafe(yield_queue.get(), loop),
        )
        asyncio_getter = asyncio.wrap_future(getter)
        item = await asyncio_getter
        if isinstance(item, _DoneSingleton):
            break
        yield item
    # if there was an exception then this should re-raise it in the calling loop
    _ = fut.result()


WrappedObj = TypeVar("WrappedObj", bound=AsyncioConfigurable, covariant=True)


class CallBridger(Generic[WrappedObj]):
    def __init__(
        self, wrapped_obj: WrappedObj, loop: asyncio.AbstractEventLoop
    ) -> None:
        self.wrapped_obj = wrapped_obj
        self._loop = loop

    def __getattribute__(self, attr_name: str) -> Any:
        # Almost every attribute retrieved from us will be for people actually
        # looking for an attribute of the managed object, so check there first.
        managed_obj = object.__getattribute__(self, "wrapped_obj")
        loop = object.__getattribute__(self, "_loop")
        try:
            attr = getattr(managed_obj, attr_name)
        except AttributeError:
            # Maybe this actually was for us? Let’s find it
            return object.__getattribute__(self, attr_name)

        if asyncio.iscoroutinefunction(attr):
            # Return coroutine result of async function
            # executed in managed thread to calling thread

            @functools.wraps(attr)
            async def wrapper(
                *args: Sequence[Any], **kwargs: Mapping[str, Any]
            ) -> WrappedReturn:
                return await call_coroutine_threadsafe(loop, attr, *args, **kwargs)

            return wrapper

        elif asyncio.iscoroutine(attr):
            # Return awaitable coroutine properties run in managed thread/loop
            fut = asyncio.run_coroutine_threadsafe(attr, loop)
            wrapped = asyncio.wrap_future(fut)
            return wrapped

        elif inspect.isasyncgenfunction(attr):
            # Return a wrapper that will exectue the resulting async generator
            # in managed thread loop

            @functools.wraps(attr)
            async def wrapper(
                *args: Sequence[Any], **kwargs: Mapping[str, Any]
            ) -> AsyncGenerator[WrappedYield, None]:
                item: WrappedYield
                async for item in execute_asyncgen_threadsafe(
                    loop, attr, *args, **kwargs
                ):
                    yield item

            return wrapper

        return attr


# TODO: BC 2020-02-25 instead of overwriting __get_attribute__ in this class
# use inspect.getmembers to iterate over appropriate members of adapted
# instance and setattr on the outer instance with the proper threadsafe
# resolution logic injected. This approach avoids requiring calls to
# object.__get_attribute__(self,...) to opt out of the overwritten
# functionality. It is more readable and protected from
# unintentional recursion.
class ThreadManager(Generic[WrappedObj]):
    """A wrapper to make every call into :py:class:`.hardware_control.API`
    execute within the same thread.

    This class spawns a worker thread and starts an event loop within.
    It then calls the async builder parameter within that worker thread's
    event loop passing thru all args and kwargs and injecting the worker
    thread's loop as a kwarg to the builder. The resulting built object
    is stored as a member of the class, and a synchronous interface to
    the managed object's members is also exposed for convenience.

    If you want to wait for the managed object's creation separately
    (with managed_thread_ready_blocking or managed_thread_ready_async)
    then pass threadmanager_nonblocking=True as a kwarg

    Example
    -------
    .. code-block::
    >>> from opentrons.hardware_control import API, ThreadManager
    >>> api_single_thread = ThreadManager(API.build_hardware_simulator)
    >>> await api_single_thread.home() # call as awaitable async
    >>> api_single_thread.sync.home() # call as blocking sync
    """

    def __init__(
        self,
        builder: Callable[..., Awaitable[WrappedObj]],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Build the ThreadManager.

        :param builder: The API function to use
        """

        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self.managed_obj: Optional[WrappedObj] = None
        self.bridged_obj: Optional[CallBridger[WrappedObj]] = None
        self._sync_managed_obj: Optional[SynchronousAdapter[WrappedObj]] = None
        is_running = threading.Event()
        self._is_running = is_running
        self._cached_modules: weakref.WeakKeyDictionary[
            AbstractModule, CallBridger[AbstractModule]
        ] = weakref.WeakKeyDictionary()
        # TODO: remove this if we switch to python 3.8
        # https://docs.python.org/3/library/asyncio-subprocess.html#subprocess-and-threads
        # On windows, the event loop and system interface is different and
        # this won't work.
        try:
            asyncio.get_child_watcher()
        except NotImplementedError:
            pass
        blocking = not kwargs.pop("threadmanager_nonblocking", False)
        target = object.__getattribute__(self, "_build_and_start_loop")
        thread = threading.Thread(
            target=target,
            name="ManagedThread",
            args=(builder, *args),
            kwargs=kwargs,
            daemon=True,
        )
        self._thread = thread
        thread.start()
        if blocking:
            object.__getattribute__(self, "managed_thread_ready_blocking")()

    def managed_thread_ready_blocking(self) -> None:
        object.__getattribute__(self, "_is_running").wait()
        if not object.__getattribute__(self, "managed_obj"):
            raise ThreadManagerException("Failed to create Managed Object")

    async def managed_thread_ready_async(self) -> None:
        is_running = object.__getattribute__(self, "_is_running")
        while not is_running.is_set():
            await asyncio.sleep(0.1)
        # Thread initialization is done.
        if not object.__getattribute__(self, "managed_obj"):
            raise ThreadManagerException("Failed to create Managed Object")

    def _build_and_start_loop(
        self,
        builder: Callable[..., Awaitable[WrappedObj]],
        *args: Sequence[Any],
        **kwargs: Mapping[str, Any],
    ) -> None:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        self._loop = loop
        try:
            managed_obj = loop.run_until_complete(builder(*args, loop=loop, **kwargs))
            self.managed_obj = managed_obj
            self.bridged_obj = CallBridger(managed_obj, loop)
            self._sync_managed_obj = SynchronousAdapter(managed_obj)
        except Exception:
            MODULE_LOG.exception("Exception in Thread Manager build")
        finally:
            object.__getattribute__(self, "_is_running").set()
            loop.run_forever()
            loop.close()

    @property
    def sync(self) -> SynchronousAdapter[WrappedObj]:
        # Why the ignore?
        # While self._sync_managed_obj is initialized None, a failure to build
        # the managed_obj and _sync_managed_obj is a catastrophic failure.
        # All callers of this property assume it to be valid.
        return self._sync_managed_obj  # type: ignore

    def __repr__(self) -> str:
        return "<ThreadManager>"

    def clean_up_tm(self) -> None:
        try:
            loop = object.__getattribute__(self, "_loop")
            loop.call_soon_threadsafe(loop.stop)
        except Exception:
            pass
        object.__setattr__(self, "_cached_modules", weakref.WeakKeyDictionary({}))
        object.__getattribute__(self, "_thread").join()

    def wrap_module(self, module: AbstractModule) -> CallBridger[AbstractModule]:
        """Return the module object wrapped in a CallBridger and cache it.

        The wrapped module objects are cached in `self._cached_modules` so they can be
        re-used throughout the module object's life, as creating a wrapper is expensive.
        We use a WeakKeyDictionary for caching so that module objects can be
        garbage collected when modules are detached (since entries in WeakKeyDictionary
        get discarded when there is no longer a strong reference to the key).
        """
        wrapper_cache = object.__getattribute__(self, "_cached_modules")
        this_module_wrapper = wrapper_cache.get(module)

        if this_module_wrapper is None:
            this_module_wrapper = CallBridger(
                module, object.__getattribute__(self, "_loop")
            )
            wrapper_cache.update({module: this_module_wrapper})

        return this_module_wrapper  # type: ignore

    def __getattribute__(self, attr_name: str) -> Any:
        # hardware_control.api.API.attached_modules is the only hardware
        # API method that returns something other than data. The module
        # objects it returns have associated methods that can be called.
        # That means they need the same wrapping treatment as the API
        # itself.
        if attr_name == "attached_modules":
            wrap = object.__getattribute__(self, "wrap_module")
            managed = object.__getattribute__(self, "managed_obj")
            attr = getattr(managed, attr_name)
            return [wrap(mod) for mod in attr]
        elif attr_name == "clean_up":
            # the wrapped object probably has this attr as well as us, and we
            # want to call both, with the wrapped one first

            # we only want to call cleanup once, and then only if the loop
            # is running
            wrapped_loop = object.__getattribute__(self, "_loop")
            if not wrapped_loop.is_running():
                return lambda: None

            wrapped_cleanup = getattr(
                object.__getattribute__(self, "bridged_obj"), "clean_up"
            )
            our_cleanup = object.__getattribute__(self, "clean_up_tm")

            def call_both() -> None:
                # the wrapped cleanup wants to happen in the managed thread,
                # started from the managed loop. our cleanup wants to happen
                # in the current thread, _after_ the wrapped cleanup is done
                # so cancelled tasks can have a chance to complete.
                async def clean_and_notify() -> None:
                    await wrapped_cleanup()
                    # this sleep allows the wrapped loop to spin to clean up the
                    # tasks we just cancelled.
                    await asyncio.sleep(0)

                fut = asyncio.run_coroutine_threadsafe(clean_and_notify(), wrapped_loop)
                fut.result()
                our_cleanup()

            return call_both

        else:
            try:
                return getattr(object.__getattribute__(self, "bridged_obj"), attr_name)
            except AttributeError:
                return object.__getattribute__(self, attr_name)

    def wrapped(self) -> WrappedObj:
        """Expose the type of the underlying wrapped object.

        This isn't a method that does anything (it just returns self again) but the cast
        means that the type of self will be what the threadmanager's generic wrapped
        object is. You can therefore use this to get typechecking when using
        ThreadManagers.

        While the generic type is what you say it is when you annotate the instance
        variable containing a ThreadManager, if you restrict yourself to annotating
         those instances using a protocol from hardware_api.protocols, things will more
        or less work out through the rest of the system. Not perfect, but ok.
        """
        return cast(WrappedObj, self)

    def wraps_instance(self, of_type: Type[Any]) -> bool:
        """Do isinstance() on the wrapped object."""
        return isinstance(object.__getattribute__(self, "managed_obj"), of_type)



"""Command models to wait for heating a Thermocycler's block."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from opentrons.protocol_engine.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler


WaitForBlockTemperatureCommandType = Literal["thermocycler/waitForBlockTemperature"]


class WaitForBlockTemperatureParams(BaseModel):
    """Input parameters to wait for Thermocycler's target block temperature."""

    moduleId: str = Field(..., description="Unique ID of the Thermocycler Module.")


class WaitForBlockTemperatureResult(BaseModel):
    """Result data from wait for Thermocycler's target block temperature."""


class WaitForBlockTemperatureImpl(
    AbstractCommandImpl[
        WaitForBlockTemperatureParams,
        WaitForBlockTemperatureResult,
    ]
):
    """Execution implementation of Thermocycler's wait for block temperature command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self,
        params: WaitForBlockTemperatureParams,
    ) -> WaitForBlockTemperatureResult:
        """Wait for a Thermocycler's target block temperature."""
        thermocycler_state = self._state_view.modules.get_thermocycler_module_substate(
            params.moduleId
        )

        # Raises error if no target temperature
        thermocycler_state.get_target_block_temperature()

        thermocycler_hardware = self._equipment.get_module_hardware_api(
            thermocycler_state.module_id
        )

        if thermocycler_hardware is not None:
            await thermocycler_hardware.wait_for_block_target()

        return WaitForBlockTemperatureResult()


class WaitForBlockTemperature(
    BaseCommand[WaitForBlockTemperatureParams, WaitForBlockTemperatureResult]
):
    """A command to wait for a Thermocycler's target block temperature."""

    commandType: WaitForBlockTemperatureCommandType = (
        "thermocycler/waitForBlockTemperature"
    )
    params: WaitForBlockTemperatureParams
    result: Optional[WaitForBlockTemperatureResult]

    _ImplementationCls: Type[WaitForBlockTemperatureImpl] = WaitForBlockTemperatureImpl


class WaitForBlockTemperatureCreate(BaseCommandCreate[WaitForBlockTemperatureParams]):
    """A request to create Thermocycler's wait for block temperature command."""

    commandType: WaitForBlockTemperatureCommandType = (
        "thermocycler/waitForBlockTemperature"
    )
    params: WaitForBlockTemperatureParams

    _CommandCls: Type[WaitForBlockTemperature] = WaitForBlockTemperature



from __future__ import annotations
import logging

from opentrons_shared_data.gripper import (
    load_definition,
    GripperModel,
    GripperDefinition,
    GripForceProfile,
)

log = logging.getLogger(__name__)

DEFAULT_GRIPPER_CALIBRATION_OFFSET = [0.0, 0.0, 0.0]


def info_num_to_model(num: str) -> GripperModel:
    """Returns a GripperModel from a string in the format X.Y.
    where X is the major model and Y is the minor model
    """
    major_model = num[0]
    minor_model = num[2]
    # we provisioned the some EVT grippers as 01 and some as 10
    # DVT will now be 1.1
    # PVT will now be 1.2
    model_map = {
        "0": {"0": GripperModel.v1, "1": GripperModel.v1},
        "1": {"0": GripperModel.v1, "1": GripperModel.v1_1, "2": GripperModel.v1_2},
    }
    return model_map[major_model][minor_model]


def load(gripper_model: GripperModel) -> GripperDefinition:
    return load_definition(gripper_model)


def duty_cycle_by_force(newton: float, profile: GripForceProfile) -> float:
    """
    Takes a force in newton and a sequence representing the polymomial
    equation of the gripper's force function in terms of duty cycle, where the
    integer represent the degree of the indeterminate (duty cycle), and
    the float representing its constant coefficient.

    The values come from shared-data/gripper/definitions/{schemaVersion}/{model}.json.

    :return: the duty-cycle value for the specified force
    """
    if profile.min <= newton <= profile.max:
        return sum(ele[1] * (newton ** ele[0]) for ele in profile.polynomial)
    else:
        raise ValueError("Gripper force out of bounds")



"""Command models to open a Thermocycler's lid."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate
from opentrons.protocol_engine.types import MotorAxis

if TYPE_CHECKING:
    from opentrons.protocol_engine.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler, MovementHandler


OpenLidCommandType = Literal["thermocycler/openLid"]


class OpenLidParams(BaseModel):
    """Input parameters to open a Thermocycler's lid."""

    moduleId: str = Field(..., description="Unique ID of the Thermocycler.")


class OpenLidResult(BaseModel):
    """Result data from opening a Thermocycler's lid."""


class OpenLidImpl(AbstractCommandImpl[OpenLidParams, OpenLidResult]):
    """Execution implementation of a Thermocycler's open lid command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        movement: MovementHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._movement = movement

    async def execute(self, params: OpenLidParams) -> OpenLidResult:
        """Open a Thermocycler's lid."""
        thermocycler_state = self._state_view.modules.get_thermocycler_module_substate(
            params.moduleId
        )
        thermocycler_hardware = self._equipment.get_module_hardware_api(
            thermocycler_state.module_id
        )

        # move the pipettes and gantry over the trash
        # do not home plunger axes because pipettes may be holding liquid
        axes_to_home = [
            MotorAxis.X,
            MotorAxis.Y,
        ] + self._state_view.motion.get_robot_mount_axes()
        await self._movement.home(axes=axes_to_home)

        if thermocycler_hardware is not None:
            await thermocycler_hardware.open()

        return OpenLidResult()


class OpenLid(BaseCommand[OpenLidParams, OpenLidResult]):
    """A command to open a Thermocycler's lid."""

    commandType: OpenLidCommandType = "thermocycler/openLid"
    params: OpenLidParams
    result: Optional[OpenLidResult]

    _ImplementationCls: Type[OpenLidImpl] = OpenLidImpl


class OpenLidCreate(BaseCommandCreate[OpenLidParams]):
    """A request to open a Thermocycler's lid."""

    commandType: OpenLidCommandType = "thermocycler/openLid"
    params: OpenLidParams

    _CommandCls: Type[OpenLid] = OpenLid



"""Translation of JSON protocol commands into ProtocolEngine commands."""
from typing import cast, List, Union
from pydantic import parse_obj_as

from opentrons_shared_data.pipette.dev_types import PipetteNameType
from opentrons_shared_data.protocol.models import (
    ProtocolSchemaV6,
    protocol_schema_v6,
    ProtocolSchemaV7,
    protocol_schema_v7,
    ProtocolSchemaV8,
    protocol_schema_v8,
)
from opentrons_shared_data import command as command_schema

from opentrons.types import MountType
from opentrons.protocol_engine import (
    commands as pe_commands,
    LabwareLocation,
    ModuleModel,
    DeckSlotLocation,
    Liquid,
)
from opentrons.protocol_engine.types import HexColor


class CommandTranslatorError(Exception):
    """An error raised to indicate an internal programmer error."""

    pass


def _translate_labware_command(
    protocol: ProtocolSchemaV6,
    command: protocol_schema_v6.Command,
) -> pe_commands.LoadLabwareCreate:
    labware_id = command.params.labwareId
    # v6 data model supports all commands and therefor most props are optional.
    # load labware command must contain labware_id and definition_id.
    assert labware_id is not None
    definition_id = protocol.labware[labware_id].definitionId
    assert definition_id is not None
    labware_command = pe_commands.LoadLabwareCreate(
        params=pe_commands.LoadLabwareParams(
            labwareId=command.params.labwareId,
            displayName=protocol.labware[labware_id].displayName,
            version=protocol.labwareDefinitions[definition_id].version,
            namespace=protocol.labwareDefinitions[definition_id].namespace,
            loadName=protocol.labwareDefinitions[definition_id].parameters.loadName,
            location=parse_obj_as(
                # https://github.com/samuelcolvin/pydantic/issues/1847
                LabwareLocation,  # type: ignore[arg-type]
                command.params.location,
            ),
        ),
        key=command.key,
    )
    return labware_command


def _translate_v7_labware_command(
    command: protocol_schema_v7.Command,
) -> pe_commands.LoadLabwareCreate:
    labware_id = command.params.labwareId
    # v7 data model supports all commands and therefor most props are optional.
    # load labware command must contain labware_id and definition_id.
    assert labware_id is not None
    assert command.params.version is not None
    assert command.params.namespace is not None
    assert command.params.loadName is not None

    labware_command = pe_commands.LoadLabwareCreate(
        params=pe_commands.LoadLabwareParams(
            labwareId=command.params.labwareId,
            displayName=command.params.displayName,
            version=command.params.version,
            namespace=command.params.namespace,
            loadName=command.params.loadName,
            location=parse_obj_as(
                # https://github.com/samuelcolvin/pydantic/issues/1847
                LabwareLocation,  # type: ignore[arg-type]
                command.params.location,
            ),
        ),
        key=command.key,
    )
    return labware_command


def _translate_module_command(
    protocol: ProtocolSchemaV6,
    command: protocol_schema_v6.Command,
) -> pe_commands.CommandCreate:
    module_id = command.params.moduleId
    modules = protocol.modules
    # v6 data model supports all commands and therefor most props are optional.
    # load module command must contain module_id. modules cannot be None.
    assert module_id is not None
    assert modules is not None
    translated_obj = pe_commands.LoadModuleCreate(
        params=pe_commands.LoadModuleParams(
            model=ModuleModel(modules[module_id].model),
            location=DeckSlotLocation.parse_obj(command.params.location),
            moduleId=command.params.moduleId,
        ),
        key=command.key,
    )
    return translated_obj


def _translate_v7_module_command(
    command: protocol_schema_v7.Command,
) -> pe_commands.CommandCreate:
    module_id = command.params.moduleId
    # v7 data model supports all commands and therefor most props are optional.
    # load module command must contain module_id. modules cannot be None.
    assert module_id is not None
    assert command.params.model is not None
    translated_obj = pe_commands.LoadModuleCreate(
        params=pe_commands.LoadModuleParams(
            model=ModuleModel(command.params.model),
            location=DeckSlotLocation.parse_obj(command.params.location),
            moduleId=command.params.moduleId,
        ),
        key=command.key,
    )
    return translated_obj


def _translate_pipette_command(
    protocol: ProtocolSchemaV6,
    command: protocol_schema_v6.Command,
) -> pe_commands.CommandCreate:
    pipette_id = command.params.pipetteId
    # v6 data model supports all commands and therefor most props are optional.
    # load pipette command must contain pipette_id.
    assert pipette_id is not None
    translated_obj = pe_commands.LoadPipetteCreate(
        params=pe_commands.LoadPipetteParams(
            pipetteName=PipetteNameType(protocol.pipettes[pipette_id].name),
            mount=MountType(command.params.mount),
            pipetteId=command.params.pipetteId,
        ),
        key=command.key,
    )
    return translated_obj


def _translate_v7_pipette_command(
    command: protocol_schema_v7.Command,
) -> pe_commands.CommandCreate:
    pipette_id = command.params.pipetteId
    # v7 data model supports all commands and therefor most props are optional.
    # load pipette command must contain pipette_id.
    assert pipette_id is not None
    assert command.params.pipetteName is not None
    translated_obj = pe_commands.LoadPipetteCreate(
        params=pe_commands.LoadPipetteParams(
            pipetteName=PipetteNameType(command.params.pipetteName),
            mount=MountType(command.params.mount),
            pipetteId=command.params.pipetteId,
        ),
        key=command.key,
    )
    return translated_obj


def _translate_simple_command(
    command: Union[
        protocol_schema_v6.Command,
        protocol_schema_v7.Command,
        protocol_schema_v8.Command,
    ]
) -> pe_commands.CommandCreate:
    dict_command = command.dict(exclude_none=True)

    # map deprecated `delay` commands to `waitForResume` / `waitForDuration`
    if dict_command["commandType"] == "delay":
        if "waitForResume" in dict_command["params"]:
            dict_command["commandType"] = "waitForResume"
        else:
            dict_command["commandType"] = "waitForDuration"

    translated_obj = cast(
        pe_commands.CommandCreate,
        parse_obj_as(
            # https://github.com/samuelcolvin/pydantic/issues/1847
            pe_commands.CommandCreate,  # type: ignore[arg-type]
            dict_command,
        ),
    )
    return translated_obj


class JsonTranslator:
    """Class that translates commands/liquids from PD/JSON to ProtocolEngine."""

    def translate_liquids(
        self, protocol: Union[ProtocolSchemaV6, ProtocolSchemaV7]
    ) -> List[Liquid]:
        """Takes json protocol v6 and translates liquids->protocol engine liquids."""
        protocol_liquids = protocol.liquids or {}

        return [
            Liquid(
                id=liquid_id,
                displayName=liquid.displayName,
                description=liquid.description,
                displayColor=HexColor(__root__=liquid.displayColor)
                if liquid.displayColor is not None
                else None,
            )
            for liquid_id, liquid in protocol_liquids.items()
        ]

    def translate_commands(
        self,
        protocol: Union[ProtocolSchemaV8, ProtocolSchemaV7, ProtocolSchemaV6],
    ) -> List[pe_commands.CommandCreate]:
        """Takes json protocol and translates commands->protocol engine commands."""
        if isinstance(protocol, ProtocolSchemaV6):
            return self._translate_v6_commands(protocol)
        elif isinstance(protocol, ProtocolSchemaV7):
            return self._translate_v7_commands(protocol)
        else:
            return self._translate_v8_commands(protocol)

    def _translate_v6_commands(
        self,
        protocol: ProtocolSchemaV6,
    ) -> List[pe_commands.CommandCreate]:
        """Takes json protocol v6 and translates commands->protocol engine commands."""
        commands_list: List[pe_commands.CommandCreate] = []
        for command in protocol.commands:
            if command.commandType == "loadPipette":
                translated_obj = _translate_pipette_command(protocol, command)
            elif command.commandType == "loadModule":
                translated_obj = _translate_module_command(protocol, command)
            elif command.commandType == "loadLabware":
                translated_obj = _translate_labware_command(protocol, command)
            else:
                translated_obj = _translate_simple_command(command)
            commands_list.append(translated_obj)
        return commands_list

    def _translate_v7_commands(
        self,
        protocol: Union[ProtocolSchemaV7],
    ) -> List[pe_commands.CommandCreate]:
        """Takes json protocol v7 and translates commands->protocol engine commands."""
        commands_list: List[pe_commands.CommandCreate] = []
        for command in protocol.commands:
            translated_obj = _translate_simple_command(command)
            commands_list.append(translated_obj)
        return commands_list

    def _translate_v8_commands(
        self, protocol: ProtocolSchemaV8
    ) -> List[pe_commands.CommandCreate]:
        """Translate commands in json protocol schema v8, which might be of different command schemas."""
        command_schema_ref = protocol.commandSchemaId
        # these calls will raise if the command schema version is invalid or unknown
        command_schema_version = command_schema.schema_version_from_ref(
            command_schema_ref
        )
        command_schema_string = command_schema.load_schema_string(  # noqa: F841
            command_schema_version
        )

        return [_translate_simple_command(command) for command in protocol.commands]



"""Simulating AbstractRunner factory."""

from opentrons.config import feature_flags
from opentrons.hardware_control import API as OT2API, HardwareControlAPI
from opentrons.protocols.api_support import deck_type
from opentrons.protocols.api_support.deck_type import should_load_fixed_trash
from opentrons.protocol_engine import (
    Config as ProtocolEngineConfig,
    DeckType,
    create_protocol_engine,
)
from opentrons.protocol_reader.protocol_source import ProtocolConfig

from opentrons_shared_data.robot.dev_types import RobotType

from .legacy_wrappers import LegacySimulatingContextCreator
from .protocol_runner import create_protocol_runner, AbstractRunner


async def create_simulating_runner(
    robot_type: RobotType, protocol_config: ProtocolConfig
) -> AbstractRunner:
    """Create a AbstractRunner wired to a simulating HardwareControlAPI.

    Example:
        ```python
        from pathlib import Path
        from typing import List
        from opentrons.protocol_engine import Command
        from opentrons.protocol_runner import (
            ProtocolType,
            ProtocolFile,
            AbstractRunner,
            create_simulating_runner,
        )

        protocol = ProtocolFile(
            protocol_type=ProtocolType.PYTHON,
            files=[Path("/path/to/protocol.py")],
        )
        runner: AbstractRunner = await create_simulating_runner()
        commands: List[Command] = await runner.run(protocol)
        ```
    """
    simulating_hardware_api = await _build_hardware_simulator_for_robot_type(
        robot_type=robot_type
    )

    # TODO(mc, 2021-08-25): move initial home to protocol engine
    await simulating_hardware_api.home()

    protocol_engine = await create_protocol_engine(
        hardware_api=simulating_hardware_api,
        config=ProtocolEngineConfig(
            robot_type=robot_type,
            deck_type=DeckType(deck_type.for_simulation(robot_type)),
            ignore_pause=True,
            use_virtual_modules=True,
            use_virtual_gripper=True,
            use_simulated_deck_config=True,
            use_virtual_pipettes=(not feature_flags.disable_fast_protocol_upload()),
        ),
        load_fixed_trash=should_load_fixed_trash(protocol_config),
    )

    simulating_legacy_context_creator = LegacySimulatingContextCreator(
        hardware_api=simulating_hardware_api,
        protocol_engine=protocol_engine,
    )

    return create_protocol_runner(
        protocol_config=protocol_config,
        protocol_engine=protocol_engine,
        hardware_api=simulating_hardware_api,
        legacy_context_creator=simulating_legacy_context_creator,
    )


async def _build_hardware_simulator_for_robot_type(
    robot_type: RobotType,
) -> HardwareControlAPI:
    if robot_type == "OT-2 Standard":
        return await OT2API.build_hardware_simulator()
    elif robot_type == "OT-3 Standard":
        # Inline import because OT3API is not present to import on an OT-2 system.
        from opentrons.hardware_control.ot3api import OT3API

        return await OT3API.build_hardware_simulator()



import logging
import functools

from typing import Any, Dict, Optional, Set, Tuple, Union, cast

from opentrons.types import Point

from opentrons_shared_data.pipette.pipette_definition import (
    PipetteConfigurations,
    PlungerPositions,
    MotorConfigurations,
    SupportedTipsDefinition,
    PickUpTipConfigurations,
    PressFitPickUpTipConfiguration,
    CamActionPickUpTipConfiguration,
    DropTipConfigurations,
    PlungerHomingConfigurations,
    PipetteNameType,
    PipetteModelVersionType,
    PipetteLiquidPropertiesDefinition,
    default_tip_for_liquid_class,
)
from opentrons_shared_data.errors.exceptions import (
    InvalidLiquidClassName,
    CommandPreconditionViolated,
    PythonException,
)
from ..instrument_abc import AbstractInstrument
from ..instrument_helpers import (
    piecewise_volume_conversion,
    PIPETTING_FUNCTION_FALLBACK_VERSION,
    PIPETTING_FUNCTION_LATEST_VERSION,
)
from .instrument_calibration import (
    save_pipette_offset_calibration,
    load_pipette_offset,
    PipetteOffsetByPipetteMount,
)
from opentrons_shared_data.pipette.dev_types import (
    UlPerMmAction,
    PipetteName,
    PipetteModel,
)
from opentrons_shared_data.pipette import (
    load_data as load_pipette_data,
    types as pip_types,
)
from opentrons.hardware_control.types import CriticalPoint, OT3Mount
from opentrons.hardware_control.errors import InvalidCriticalPoint
from opentrons.hardware_control import nozzle_manager

mod_log = logging.getLogger(__name__)


class Pipette(AbstractInstrument[PipetteConfigurations]):
    """A class to gather and track pipette state and configs.

    This class should not touch hardware or call back out to the hardware
    control API. Its only purpose is to gather state.
    """

    DictType = Dict[str, Union[str, float, bool]]
    #: The type of this data class as a dict

    def __init__(
        self,
        config: PipetteConfigurations,
        pipette_offset_cal: PipetteOffsetByPipetteMount,
        pipette_id: Optional[str] = None,
        use_old_aspiration_functions: bool = False,
    ) -> None:
        self._config = config
        self._config_as_dict = config.dict()
        self._plunger_motor_current = config.plunger_motor_configurations
        self._pick_up_configurations = config.pick_up_tip_configurations
        self._plunger_homing_configurations = config.plunger_homing_configurations
        self._drop_configurations = config.drop_tip_configurations
        self._pipette_offset = pipette_offset_cal
        self._pipette_type = self._config.pipette_type
        self._pipette_version = self._config.version
        self._max_channels = self._config.channels
        self._backlash_distance = config.backlash_distance

        self._liquid_class_name = pip_types.LiquidClasses.default
        self._liquid_class = self._config.liquid_properties[self._liquid_class_name]

        # TODO (lc 12-05-2022) figure out how we can safely deprecate "name" and "model"
        self._pipette_name = PipetteNameType(
            pipette_type=config.pipette_type,
            pipette_channels=config.channels,
            pipette_generation=config.display_category,
        )
        self._acting_as = self._pipette_name
        self._pipette_model = PipetteModelVersionType(
            pipette_type=config.pipette_type,
            pipette_channels=config.channels,
            pipette_version=config.version,
        )
        self._nozzle_offset = self._config.nozzle_offset
        self._nozzle_manager = (
            nozzle_manager.NozzleConfigurationManager.build_from_config(self._config)
        )
        self._current_volume = 0.0
        self._working_volume = float(self._liquid_class.max_volume)
        self._current_tip_length = 0.0
        self._has_tip_length: Optional[bool] = None
        self._current_tiprack_diameter = 0.0
        self._pipette_id = pipette_id
        self._log = mod_log.getChild(
            self._pipette_id if self._pipette_id else "<unknown>"
        )
        self._log.info(
            "loaded: {}, pipette offset: {}".format(
                self._pipette_model, self._pipette_offset.offset
            )
        )
        self.ready_to_aspirate = False

        self._active_tip_setting_name = default_tip_for_liquid_class(self._liquid_class)
        self._active_tip_settings = self._liquid_class.supported_tips[
            self._active_tip_setting_name
        ]
        self._fallback_tip_length = self._active_tip_settings.default_tip_length

        self._aspirate_flow_rate = (
            self._active_tip_settings.default_aspirate_flowrate.default
        )
        self._dispense_flow_rate = (
            self._active_tip_settings.default_dispense_flowrate.default
        )
        self._blow_out_flow_rate = (
            self._active_tip_settings.default_blowout_flowrate.default
        )
        self._flow_acceleration = self._active_tip_settings.default_flow_acceleration

        self._tip_overlap_lookup = self._liquid_class.tip_overlap_dictionary

        if use_old_aspiration_functions:
            self._pipetting_function_version = PIPETTING_FUNCTION_FALLBACK_VERSION
        else:
            self._pipetting_function_version = PIPETTING_FUNCTION_LATEST_VERSION

    @property
    def config(self) -> PipetteConfigurations:
        return self._config

    @property
    def liquid_class(self) -> PipetteLiquidPropertiesDefinition:
        return self._liquid_class

    @property
    def liquid_class_name(self) -> pip_types.LiquidClasses:
        return self._liquid_class_name

    @property
    def channels(self) -> pip_types.PipetteChannelType:
        return self._max_channels

    @property
    def backlash_distance(self) -> float:
        return self._backlash_distance

    @property
    def tip_overlap(self) -> Dict[str, float]:
        return self._tip_overlap_lookup

    @property
    def nozzle_offset(self) -> Point:
        return self._nozzle_manager.starting_nozzle_offset

    @property
    def nozzle_manager(self) -> nozzle_manager.NozzleConfigurationManager:
        return self._nozzle_manager

    @property
    def pipette_offset(self) -> PipetteOffsetByPipetteMount:
        return self._pipette_offset

    @property
    def plunger_positions(self) -> PlungerPositions:
        return self._config.plunger_positions_configurations[self._liquid_class_name]

    @property
    def plunger_motor_current(self) -> MotorConfigurations:
        return self._plunger_motor_current

    @property
    def pick_up_configurations(self) -> PickUpTipConfigurations:
        return self._pick_up_configurations

    @pick_up_configurations.setter
    def pick_up_configurations(self, pick_up_configs: PickUpTipConfigurations) -> None:
        self._pick_up_configurations = pick_up_configs

    @property
    def plunger_homing_configurations(self) -> PlungerHomingConfigurations:
        return self._plunger_homing_configurations

    @property
    def drop_configurations(self) -> DropTipConfigurations:
        return self._drop_configurations

    @property
    def active_tip_settings(self) -> SupportedTipsDefinition:
        return self._active_tip_settings

    @property
    def push_out_volume(self) -> float:
        return self._active_tip_settings.default_push_out_volume

    def act_as(self, name: PipetteName) -> None:
        """Reconfigure to act as ``name``. ``name`` must be either the
        actual name of the pipette, or a name in its back-compatibility
        config.
        """
        raise NotImplementedError(
            "Backwards compatibility is not supported at this time."
        )

    def update_config_item(self, elem_name: Dict[str, Any]) -> None:
        raise NotImplementedError("Update config is not supported at this time.")

    @property
    def acting_as(self) -> PipetteName:
        return cast(PipetteName, f"{self._acting_as}")

    def reload_configurations(self) -> None:
        self._config = load_pipette_data.load_definition(
            self._pipette_model.pipette_type,
            self._pipette_model.pipette_channels,
            self._pipette_model.pipette_version,
        )
        self._config_as_dict = self._config.dict()

    def reset_state(self) -> None:
        self._current_volume = 0.0
        self._working_volume = float(self.liquid_class.max_volume)
        self._current_tip_length = 0.0
        self._has_tip_length = None
        self._current_tiprack_diameter = 0.0
        self.ready_to_aspirate = False
        #: True if ready to aspirate
        self.set_liquid_class_by_name("default")
        self.set_tip_type(default_tip_for_liquid_class(self._liquid_class))

        self._aspirate_flow_rate = (
            self._active_tip_settings.default_aspirate_flowrate.default
        )
        self._dispense_flow_rate = (
            self._active_tip_settings.default_dispense_flowrate.default
        )
        self._blow_out_flow_rate = (
            self._active_tip_settings.default_blowout_flowrate.default
        )
        self._flow_acceleration = self._active_tip_settings.default_flow_acceleration

        self._tip_overlap_lookup = self.liquid_class.tip_overlap_dictionary
        self._nozzle_manager = (
            nozzle_manager.NozzleConfigurationManager.build_from_config(self._config)
        )

    def reset_pipette_offset(self, mount: OT3Mount, to_default: bool) -> None:
        """Reset the pipette offset to system defaults."""
        if to_default:
            self._pipette_offset = load_pipette_offset(pip_id=None, mount=mount)
        else:
            self._pipette_offset = load_pipette_offset(self._pipette_id, mount)

    def save_pipette_offset(
        self, mount: OT3Mount, offset: Point
    ) -> PipetteOffsetByPipetteMount:
        """Update the pipette offset to a new value."""
        save_pipette_offset_calibration(self._pipette_id, mount, offset)
        self._pipette_offset = load_pipette_offset(self._pipette_id, mount)
        return self._pipette_offset

    @property
    def name(self) -> PipetteName:
        return cast(PipetteName, f"{self._pipette_name}")

    @property
    def model(self) -> PipetteModel:
        return cast(PipetteModel, f"{self._pipette_model}")

    @property
    def pipette_type(self) -> pip_types.PipetteModelType:
        return self._pipette_type

    @property
    def pipette_id(self) -> Optional[str]:
        return self._pipette_id

    def critical_point(self, cp_override: Optional[CriticalPoint] = None) -> Point:
        """
        The vector from the pipette's origin to its critical point. The
        critical point for a pipette is the end of the nozzle if no tip is
        attached, or the end of the tip if a tip is attached.

        If `cp_override` is specified and valid - so is either
        :py:attr:`CriticalPoint.NOZZLE` or :py:attr:`CriticalPoint.TIP` when
        we have a tip, or :py:attr:`CriticalPoint.XY_CENTER` - the specified
        critical point will be used.
        """
        if cp_override in [
            CriticalPoint.GRIPPER_JAW_CENTER,
            CriticalPoint.GRIPPER_FRONT_CALIBRATION_PIN,
            CriticalPoint.GRIPPER_REAR_CALIBRATION_PIN,
        ]:
            raise InvalidCriticalPoint(cp_override.name, "pipette")

        instr = Point(*self._pipette_offset.offset)
        cp_with_tip_length = self._nozzle_manager.critical_point_with_tip_length(
            cp_override,
            self.current_tip_length if cp_override != CriticalPoint.NOZZLE else 0.0,
        )
        cp = cp_with_tip_length + instr

        if self._log.isEnabledFor(logging.DEBUG):
            info_str = "cp: {}{}: {} (from: ".format(
                cp_override, " (from override)" if cp_override else "", cp
            )
            info_str += "model offset: {} + instrument offset: {}".format(
                cp_with_tip_length, instr
            )
            info_str += " - tip_length: {}".format(self.current_tip_length)
            info_str += ")"
            self._log.debug(info_str)

        return cp

    @property
    def current_volume(self) -> float:
        """The amount of liquid currently aspirated"""
        return self._current_volume

    @property
    def current_tip_length(self) -> float:
        """The length of the current tip attached (0.0 if no tip)"""
        return self._current_tip_length

    @current_tip_length.setter
    def current_tip_length(self, tip_length: float) -> None:
        self._current_tip_length = tip_length
        self._has_tip_length = True

    @property
    def current_tiprack_diameter(self) -> float:
        """The diameter of the current tip rack (0.0 if no tip)"""
        return self._current_tiprack_diameter

    @current_tiprack_diameter.setter
    def current_tiprack_diameter(self, diameter: float) -> None:
        self._current_tiprack_diameter = diameter

    @property
    def aspirate_flow_rate(self) -> float:
        """Current active flow rate (not config value)"""
        return self._aspirate_flow_rate

    @aspirate_flow_rate.setter
    def aspirate_flow_rate(self, new_flow_rate: float) -> None:
        assert new_flow_rate > 0
        self._aspirate_flow_rate = new_flow_rate

    @property
    def dispense_flow_rate(self) -> float:
        """Current active flow rate (not config value)"""
        return self._dispense_flow_rate

    @dispense_flow_rate.setter
    def dispense_flow_rate(self, new_flow_rate: float) -> None:
        assert new_flow_rate > 0
        self._dispense_flow_rate = new_flow_rate

    @property
    def blow_out_flow_rate(self) -> float:
        """Current active flow rate (not config value)"""
        return self._blow_out_flow_rate

    @blow_out_flow_rate.setter
    def blow_out_flow_rate(self, new_flow_rate: float) -> None:
        assert new_flow_rate > 0
        self._blow_out_flow_rate = new_flow_rate

    @property
    def flow_acceleration(self) -> float:
        """Current active flow acceleration (not config value)"""
        return self._flow_acceleration

    @flow_acceleration.setter
    def flow_acceleration(self, new_flow_acceleration: float) -> None:
        assert new_flow_acceleration > 0
        self._flow_acceleration = new_flow_acceleration

    @property
    def aspirate_flow_rates_lookup(self) -> Dict[str, float]:
        return self._active_tip_settings.default_aspirate_flowrate.values_by_api_level

    @property
    def dispense_flow_rates_lookup(self) -> Dict[str, float]:
        return self._active_tip_settings.default_dispense_flowrate.values_by_api_level

    @property
    def blow_out_flow_rates_lookup(self) -> Dict[str, float]:
        return self._active_tip_settings.default_blowout_flowrate.values_by_api_level

    @property
    def working_volume(self) -> float:
        """The working volume of the pipette"""
        return self._working_volume

    @working_volume.setter
    def working_volume(self, tip_volume: float) -> None:
        """The working volume is the current tip max volume"""
        self.set_tip_type_by_volume(tip_volume)
        self._working_volume = min(tip_volume, self.liquid_class.max_volume)

    @property
    def minimum_volume(self) -> float:
        """The smallest controllable volume the pipette can handle in this liquid class."""
        return self.liquid_class.min_volume

    @property
    def available_volume(self) -> float:
        """The amount of liquid possible to aspirate"""
        return self.working_volume - self.current_volume

    def set_current_volume(self, new_volume: float) -> None:
        assert new_volume >= 0
        assert new_volume <= self.working_volume
        self._current_volume = new_volume

    def add_current_volume(self, volume_incr: float) -> None:
        assert self.ok_to_add_volume(volume_incr)
        self._current_volume += volume_incr

    def remove_current_volume(self, volume_incr: float) -> None:
        assert self._current_volume >= volume_incr
        self._current_volume -= volume_incr

    def ok_to_add_volume(self, volume_incr: float) -> bool:
        return self.current_volume + volume_incr <= self.working_volume

    def ok_to_push_out(self, push_out_dist_mm: float) -> bool:
        return push_out_dist_mm <= (
            self.plunger_positions.blow_out - self.plunger_positions.bottom
        )

    def update_nozzle_configuration(
        self,
        back_left_nozzle: str,
        front_right_nozzle: str,
        starting_nozzle: Optional[str] = None,
    ) -> None:
        """
        Update nozzle configuration manager.
        """
        self._nozzle_manager.update_nozzle_configuration(
            back_left_nozzle, front_right_nozzle, starting_nozzle
        )

    def reset_nozzle_configuration(self) -> None:
        """
        Reset nozzle configuration manager.
        """
        self._nozzle_manager.reset_to_default_configuration()

    def add_tip(self, tip_length: float) -> None:
        """
        Add a tip to the pipette for position tracking and validation
        (effectively updates the pipette's critical point)

        :param tip_length: a positive, non-zero float presenting the distance
            in Z from the end of the pipette nozzle to the end of the tip
        :return:
        """
        assert tip_length > 0.0, "tip_length must be greater than 0"
        assert not self.has_tip_length
        self._current_tip_length = tip_length
        self._has_tip_length = True

    def remove_tip(self) -> None:
        """
        Remove the tip from the pipette (effectively updates the pipette's
        critical point)
        """
        self._current_tip_length = 0.0
        self._has_tip_length = False

    @property
    def has_tip(self) -> bool:
        return self.has_tip_length

    @property
    def has_tip_length(self) -> bool:
        return self.current_tip_length > 0.0

    @property
    def tip_presence_check_dist_mm(self) -> float:
        return self._config.tip_presence_check_distance_mm

    @property
    def tip_presence_responses(self) -> int:
        # TODO: put this in shared-data
        return 2 if self.channels > 8 else 1

    # Cache max is chosen somewhat arbitrarily. With a float is input we don't
    # want this to unbounded.
    @functools.lru_cache(maxsize=100)
    def ul_per_mm(self, ul: float, action: UlPerMmAction) -> float:
        if action == "aspirate":
            fallback = self._active_tip_settings.aspirate.default[
                PIPETTING_FUNCTION_FALLBACK_VERSION
            ]
            sequence = self._active_tip_settings.aspirate.default.get(
                self._pipetting_function_version, fallback
            )
        elif action == "blowout":
            return self._config.shaft_ul_per_mm
        else:
            fallback = self._active_tip_settings.dispense.default[
                PIPETTING_FUNCTION_FALLBACK_VERSION
            ]
            sequence = self._active_tip_settings.dispense.default.get(
                self._pipetting_function_version, fallback
            )
        return piecewise_volume_conversion(ul, sequence)

    def __str__(self) -> str:
        return "{} current volume {}ul critical point: {} at {}".format(
            self._config.display_name,
            self.current_volume,
            "tip end" if self.has_tip_length else "nozzle end",
            0,
        )

    def __repr__(self) -> str:
        return "<{}: {} {}>".format(
            self.__class__.__name__, self._config.display_name, id(self)
        )

    def as_dict(self) -> "Pipette.DictType":
        # TODO (lc 12-05-2022) Kill this code ASAP
        self._config_as_dict.update(
            {
                "current_volume": self.current_volume,
                "available_volume": self.available_volume,
                "name": self.name,
                "model": self.model,
                "pipette_id": self.pipette_id,
                "has_tip": self.has_tip_length,
                "working_volume": self.working_volume,
                "aspirate_flow_rate": self.aspirate_flow_rate,
                "dispense_flow_rate": self.dispense_flow_rate,
                "blow_out_flow_rate": self.blow_out_flow_rate,
                "flow_acceleration": self.flow_acceleration,
                "default_aspirate_flow_rates": self.aspirate_flow_rates_lookup,
                "default_blow_out_flow_rates": self.blow_out_flow_rates_lookup,
                "default_dispense_flow_rates": self.dispense_flow_rates_lookup,
                "default_flow_acceleration": self.active_tip_settings.default_flow_acceleration,
                "tip_length": self.current_tip_length,
                "return_tip_height": self.active_tip_settings.default_return_tip_height,
                "tip_overlap": self.tip_overlap,
                "back_compat_names": self._config.pipette_backcompat_names,
                "supported_tips": self.liquid_class.supported_tips,
            }
        )
        return self._config_as_dict

    def set_liquid_class_by_name(self, class_name: str) -> None:
        """Change the currently active liquid class."""
        if self.current_volume > 0:
            raise CommandPreconditionViolated(
                "Cannot switch liquid classes when liquid is in the tip"
            )
        try:
            new_name = pip_types.LiquidClasses[class_name]
            if new_name == self._liquid_class_name:
                return
            new_class = self._config.liquid_properties[new_name]
        except KeyError:
            raise InvalidLiquidClassName(
                message=f"Liquid class {class_name} is not valid for {self._config.display_name}",
                detail={
                    "requested-class-name": class_name,
                    "pipette-model": str(self._pipette_model),
                },
            )
        if (
            self.has_tip_length
            and self._active_tip_setting_name not in new_class.supported_tips
        ):
            raise CommandPreconditionViolated(
                message=f"Requested liquid class {class_name} does not support the currently attached tip",
                detail={
                    "requested-class-name": class_name,
                    "current-tip": str(self._active_tip_setting_name),
                },
            )
        self._liquid_class_name = new_name
        self._liquid_class = new_class
        if not self.has_tip_length:
            new_tip_class = sorted(
                [tip for tip in self._liquid_class.supported_tips.keys()],
                key=lambda tt: tt.value,
            )[0]
        else:
            new_tip_class = self._active_tip_setting_name
        self.set_tip_type(new_tip_class)

        self.ready_to_aspirate = False

    def set_tip_type_by_volume(self, tip_volume: float) -> None:
        """Change the currently active tip type."""
        intified = float(int(tip_volume))
        try:
            new_name = pip_types.PipetteTipType(int(intified))
        except (ValueError, KeyError) as e:
            raise InvalidLiquidClassName(
                message=f"There is no configuration for tips of volume {intified} in liquid class {str(self._liquid_class_name)}",
                detail={
                    "current-liquid-class": str(self._liquid_class_name),
                    "requested-volume": str(intified),
                },
                wrapping=[PythonException(e)],
            ) from e
        self.set_tip_type(new_name)

    def set_tip_type(self, tip_type: pip_types.PipetteTipType) -> None:
        """Change the currently active tip by its exact type."""
        try:
            new_tips = self._liquid_class.supported_tips[tip_type]
        except KeyError as e:
            raise InvalidLiquidClassName(
                message=f"There is no configuration for {tip_type.name} in liquid class {str(self._liquid_class_name)} on a {self._config.display_name}",
                detail={
                    "current-liquid-class": str(self._liquid_class_name),
                    "requested-type": tip_type.name,
                },
                wrapping=[PythonException(e)],
            ) from e

        self._active_tip_setting_name = tip_type
        self._active_tip_settings = new_tips

        self._aspirate_flow_rate = (
            self._active_tip_settings.default_aspirate_flowrate.default
        )
        self._dispense_flow_rate = (
            self._active_tip_settings.default_dispense_flowrate.default
        )
        self._blow_out_flow_rate = (
            self._active_tip_settings.default_blowout_flowrate.default
        )
        self._flow_acceleration = self._active_tip_settings.default_flow_acceleration

        self._fallback_tip_length = self._active_tip_settings.default_tip_length
        self._tip_overlap_lookup = self.liquid_class.tip_overlap_dictionary
        self._working_volume = min(tip_type.value, self.liquid_class.max_volume)

    def get_pick_up_configuration_for_tip_count(
        self, count: int
    ) -> Union[CamActionPickUpTipConfiguration, PressFitPickUpTipConfiguration]:
        for config in (
            self._config.pick_up_tip_configurations.press_fit,
            self._config.pick_up_tip_configurations.cam_action,
        ):
            if not config:
                continue
            if count in config.current_by_tip_count:
                return config
        raise CommandPreconditionViolated(
            message=f"No pick up tip configuration for {count} tips",
        )


def _reload_and_check_skip(
    new_config: PipetteConfigurations,
    attached_instr: Pipette,
    pipette_offset: PipetteOffsetByPipetteMount,
    use_old_aspiration_functions: bool,
) -> Tuple[Pipette, bool]:
    # Once we have determined that the new and attached pipettes
    # are similar enough that we might skip, see if the configs
    # match closely enough.
    # Returns a pipette object and True if we may skip hw reconfig
    # TODO this can potentially be removed in a follow-up refactor.
    if new_config == attached_instr.config:
        # Same config, good enough
        return attached_instr, True
    else:
        newdict = new_config.dict()
        olddict = attached_instr.config.dict()
        changed: Set[str] = set()
        for k in newdict.keys():
            if newdict[k] != olddict[k]:
                changed.add(k)
        if changed.intersection("quirks"):
            # Something has changed that requires reconfig
            p = Pipette(
                new_config,
                pipette_offset,
                attached_instr._pipette_id,
                use_old_aspiration_functions,
            )
            p.act_as(attached_instr.acting_as)
            return p, False
        # Good to skip, just need to update calibration offset and update_info
        attached_instr._pipette_offset = pipette_offset
        return attached_instr, True


def load_from_config_and_check_skip(
    config: Optional[PipetteConfigurations],
    attached: Optional[Pipette],
    requested: Optional[PipetteName],
    serial: Optional[str],
    pipette_offset: PipetteOffsetByPipetteMount,
    use_old_aspiration_functions: bool,
) -> Tuple[Optional[Pipette], bool]:
    """
    Given the pipette config for an attached pipette (if any) freshly read
    from disk, and any attached instruments,

    - Compare the new and configured pipette configs
    - Load the new configs if they differ
    - Return a bool indicating whether hardware reconfiguration may be
      skipped
    """

    if not config and not attached:
        # nothing attached now, nothing used to be attached, nothing
        # to reconfigure
        return attached, True

    if config and attached:
        # something was attached and something is attached. are they
        # the same? we can tell by comparing serials
        if serial == attached.pipette_id:
            if requested:
                # if there is an explicit instrument request, in addition
                # to checking if the old and new responses are the same
                # we also have to make sure the old pipette is properly
                # configured to the request
                if requested == attached.acting_as:
                    # similar enough to check
                    return _reload_and_check_skip(
                        config,
                        attached,
                        pipette_offset,
                        use_old_aspiration_functions,
                    )
            else:
                # if there is no request, make sure that the old pipette
                # did not have backcompat applied
                if attached.acting_as == attached.name:
                    # similar enough to check
                    return _reload_and_check_skip(
                        config,
                        attached,
                        pipette_offset,
                        use_old_aspiration_functions,
                    )

    if config:
        return (
            Pipette(config, pipette_offset, serial, use_old_aspiration_functions),
            False,
        )
    else:
        return None, False



import typing

from typing_extensions import Literal
from opentrons.hardware_control.modules.types import ModuleType
from opentrons.hardware_control.types import OT3Mount
from pydantic import BaseModel, Field, validator
from datetime import datetime

from opentrons_shared_data.pipette.dev_types import LabwareUri

from opentrons.types import Point
from opentrons.calibration_storage import types


class CalibrationStatus(BaseModel):
    markedBad: bool = False
    source: typing.Optional[types.SourceType] = None
    markedAt: typing.Optional[datetime] = None


class TipLengthModel(BaseModel):
    tipLength: float = Field(..., description="Tip length data found from calibration.")
    lastModified: datetime = Field(
        ..., description="The last time this tip length was calibrated."
    )
    uri: typing.Union[LabwareUri, Literal[""]] = Field(
        ..., description="The tiprack URI associated with the tip length data."
    )
    source: types.SourceType = Field(
        default=types.SourceType.factory, description="The source of calibration."
    )
    status: CalibrationStatus = Field(
        default_factory=CalibrationStatus,
        description="The status of the calibration data.",
    )

    @validator("tipLength")
    def ensure_tip_length_positive(cls, tipLength: float) -> float:
        if tipLength < 0.0:
            raise ValueError("Tip Length must be a positive number")
        return tipLength

    class Config:
        json_encoders = {datetime: lambda obj: obj.isoformat()}
        json_decoders = {datetime: lambda obj: datetime.fromisoformat(obj)}


class BeltCalibrationModel(BaseModel):
    attitude: types.AttitudeMatrix = Field(
        ..., description="Attitude matrix for belts found from calibration."
    )
    lastModified: datetime = Field(
        ..., description="The last time this deck was calibrated."
    )
    source: types.SourceType = Field(
        default=types.SourceType.factory, description="The source of calibration."
    )
    pipetteCalibratedWith: typing.Optional[str] = Field(
        default=None, description="The pipette id used to calibrate the deck."
    )
    status: CalibrationStatus = Field(
        default_factory=CalibrationStatus,
        description="The status of the calibration data.",
    )

    class Config:
        json_encoders = {datetime: lambda obj: obj.isoformat()}
        json_decoders = {datetime: lambda obj: datetime.fromisoformat(obj)}


class InstrumentOffsetModel(BaseModel):
    offset: Point = Field(..., description="Instrument offset found from calibration.")
    lastModified: datetime = Field(
        ..., description="The last time this instrument was calibrated."
    )
    source: types.SourceType = Field(
        default=types.SourceType.factory, description="The source of calibration."
    )
    status: CalibrationStatus = Field(
        default_factory=CalibrationStatus,
        description="The status of the calibration data.",
    )

    class Config:
        json_encoders = {datetime: lambda obj: obj.isoformat()}
        json_decoders = {datetime: lambda obj: datetime.fromisoformat(obj)}


class ModuleOffsetModel(BaseModel):
    offset: Point = Field(..., description="Module offset found from calibration.")
    mount: OT3Mount = Field(..., description="The mount used to calibrate this module.")
    slot: str = Field(
        ...,
        description=("The slot this module was calibrated in."),
    )
    module: ModuleType = Field(..., description="The module type of this module.")
    module_id: str = Field(..., description="The unique id of this module.")
    instrument_id: str = Field(
        ...,
        description="The unique id of the instrument used to calibrate this module.",
    )
    lastModified: datetime = Field(
        ..., description="The last time this module was calibrated."
    )
    source: types.SourceType = Field(
        default=types.SourceType.factory, description="The source of calibration."
    )
    status: CalibrationStatus = Field(
        default_factory=CalibrationStatus,
        description="The status of the calibration data.",
    )

    class Config:
        json_encoders = {datetime: lambda obj: obj.isoformat()}
        json_decoders = {datetime: lambda obj: datetime.fromisoformat(obj)}



"""ProtocolContext factory."""
import asyncio
from typing import Any, Dict, Optional, Union, cast

from opentrons_shared_data.labware.dev_types import LabwareDefinition

from opentrons.config import feature_flags
from opentrons.hardware_control import (
    HardwareControlAPI,
    ThreadManager,
    SynchronousAdapter,
)
from opentrons.legacy_broker import LegacyBroker
from opentrons.util.broker import Broker
from opentrons.protocol_engine import ProtocolEngine
from opentrons.protocol_engine.clients import SyncClient, ChildThreadTransport
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.deck_type import (
    should_load_fixed_trash_area_for_python_protocol,
)
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION

from .protocol_context import ProtocolContext
from .deck import Deck
from ._trash_bin import TrashBin

from .core.common import ProtocolCore as AbstractProtocolCore
from .core.legacy.deck import Deck as LegacyDeck
from .core.legacy.legacy_protocol_core import LegacyProtocolCore
from .core.legacy.labware_offset_provider import (
    AbstractLabwareOffsetProvider,
    LabwareOffsetProvider,
    NullLabwareOffsetProvider,
)
from .core.legacy_simulator.legacy_protocol_core import LegacyProtocolCoreSimulator
from .core.engine import ENGINE_CORE_API_VERSION, ProtocolCore


class ProtocolEngineCoreRequiredError(Exception):
    """Raised when a Protocol Engine core was required, but not provided.

    This can happen when creating a ProtocolContext with a high api_version.
    """


def create_protocol_context(
    api_version: APIVersion,
    *,
    hardware_api: Union[HardwareControlAPI, ThreadManager[HardwareControlAPI]],
    deck_type: str,
    protocol_engine: Optional[ProtocolEngine] = None,
    protocol_engine_loop: Optional[asyncio.AbstractEventLoop] = None,
    broker: Optional[LegacyBroker] = None,
    equipment_broker: Optional[Broker[Any]] = None,
    use_simulating_core: bool = False,
    extra_labware: Optional[Dict[str, LabwareDefinition]] = None,
    bundled_labware: Optional[Dict[str, LabwareDefinition]] = None,
    bundled_data: Optional[Dict[str, bytes]] = None,
) -> ProtocolContext:
    """Create a ProtocolContext for use in a Python protocol.

    Args:
        api_version: The API version to target.
        hardware_api: Control interface to the device's hardware.
        deck_type: What kind of deck the device has.
            This must match the deck type in `protocol_engine`'s config, if there is one.
        protocol_engine: A ProtocolEngine to use for labware offsets
            and core protocol logic. If omitted, labware offsets will
            all be (0, 0, 0) and ProtocolEngine-based core will not work.
        protocol_engine_loop: An event loop running in the thread where
            ProtocolEngine mutations must occur.
        broker: A message broker for protocol command event publishing.
        equipment_broker: A message broker for equipment load event publishing.
        use_simulating_core: For pre-ProtocolEngine API versions,
            use a simulating protocol core that will skip _most_ calls
            to the `hardware_api`.
        extra_labware: Extra labware definitions to include in
            labware definition lookup paths.
        bundled_labware: Do not use in new code. Leftover from
            experimental ZIP protocol bundles.
        bundled_data: Do not use in new code. Leftover from
            experimental ZIP protocol bundles.

    Returns:
        A ready-to-use ProtocolContext.
    """
    if api_version > MAX_SUPPORTED_VERSION:
        raise ValueError(
            f"API version {api_version} is not supported by this robot software."
            f" Please reduce your API version to {MAX_SUPPORTED_VERSION} or below"
            f" or update your robot."
        )

    sync_hardware: SynchronousAdapter[HardwareControlAPI]
    labware_offset_provider: AbstractLabwareOffsetProvider
    core: Union[ProtocolCore, LegacyProtocolCoreSimulator, LegacyProtocolCore]

    if isinstance(hardware_api, ThreadManager):
        sync_hardware = hardware_api.sync
    else:
        sync_hardware = SynchronousAdapter(hardware_api)

    if protocol_engine is not None:
        assert deck_type == protocol_engine.state_view.config.deck_type.value
        labware_offset_provider = LabwareOffsetProvider(engine=protocol_engine)
    else:
        labware_offset_provider = NullLabwareOffsetProvider()

    if api_version >= ENGINE_CORE_API_VERSION:
        # TODO(mc, 2022-8-22): replace raise with strict typing
        if protocol_engine is None or protocol_engine_loop is None:
            raise ProtocolEngineCoreRequiredError(
                "ProtocolEngine PAPI core is enabled, but no ProtocolEngine given."
            )

        engine_client_transport = ChildThreadTransport(
            engine=protocol_engine, loop=protocol_engine_loop
        )
        engine_client = SyncClient(transport=engine_client_transport)
        core = ProtocolCore(
            engine_client=engine_client,
            api_version=api_version,
            sync_hardware=sync_hardware,
        )

    # TODO(mc, 2022-8-22): remove `disable_fast_protocol_upload`
    elif use_simulating_core and not feature_flags.disable_fast_protocol_upload():
        legacy_deck = LegacyDeck(deck_type=deck_type)
        core = LegacyProtocolCoreSimulator(
            sync_hardware=sync_hardware,
            labware_offset_provider=labware_offset_provider,
            deck_layout=legacy_deck,
            equipment_broker=equipment_broker,
            api_version=api_version,
            bundled_labware=bundled_labware,
            extra_labware=extra_labware,
        )

    else:
        legacy_deck = LegacyDeck(deck_type=deck_type)
        core = LegacyProtocolCore(
            sync_hardware=sync_hardware,
            labware_offset_provider=labware_offset_provider,
            deck_layout=legacy_deck,
            equipment_broker=equipment_broker,
            api_version=api_version,
            bundled_labware=bundled_labware,
            extra_labware=extra_labware,
        )

    # TODO(mc, 2022-12-06): add API version guard in addition to instance check
    # this swap may happen once `ctx.move_labware` off-deck is implemented
    deck = None if isinstance(core, ProtocolCore) else cast(Deck, core.get_deck())

    context = ProtocolContext(
        api_version=api_version,
        # TODO(mm, 2023-05-11): This cast shouldn't be necessary.
        # Fix this by making the appropriate TypeVars covariant?
        # https://peps.python.org/pep-0484/#covariance-and-contravariance
        core=cast(AbstractProtocolCore, core),
        broker=broker,
        deck=deck,
        bundled_data=bundled_data,
    )
    # If we're loading an engine based core into the context, and we're on api level 2.16 or above, on an OT-2 we need
    # to insert a fixed trash addressable area into the protocol engine, for correctness in anything that relies on
    # knowing what addressable areas have been loaded (and any checks involving trash geometry). Because the method
    # that uses this in the core relies on the sync client and this code will run in the main thread (which if called
    # will cause a deadlock), we're directly calling the protocol engine method here where we have access to it.
    if (
        protocol_engine is not None
        and should_load_fixed_trash_area_for_python_protocol(
            api_version=api_version,
            robot_type=protocol_engine.state_view.config.robot_type,
        )
    ):
        assert isinstance(context.fixed_trash, TrashBin)
        protocol_engine.add_addressable_area(context.fixed_trash.area_name)
    return context



"""Read relevant protocol information from a set of files."""
from pathlib import Path
from typing import Optional, Sequence

from opentrons.protocols.parse import PythonParseMode

from .file_identifier import (
    FileIdentifier,
    IdentifiedFile,
    IdentifiedJsonMain,
    IdentifiedPythonMain,
    IdentifiedLabwareDefinition,
    IdentifiedData,
)
from .role_analyzer import RoleAnalyzer, RoleAnalysis
from .file_format_validator import FileFormatValidator
from .file_reader_writer import FileReaderWriter, BufferedFile
from .file_hasher import FileHasher
from .protocol_source import (
    ProtocolSource,
    ProtocolSourceFile,
    ProtocolConfig,
    ProtocolFileRole,
    JsonProtocolConfig,
    PythonProtocolConfig,
)


class ProtocolReader:
    """Collaborator to turn a set of files into a protocol object."""

    def __init__(
        self,
        file_reader_writer: Optional[FileReaderWriter] = None,
        file_identifier: Optional[FileIdentifier] = None,
        role_analyzer: Optional[RoleAnalyzer] = None,
        file_format_validator: Optional[FileFormatValidator] = None,
        file_hasher: Optional[FileHasher] = None,
    ) -> None:
        """Initialize the reader with its dependencies.

        Arguments:
            file_reader_writer: Input file reader/writer. Default impl. used if None.
            file_identifier: File identifier. Default impl. used if None.
            role_analyzer: File role analyzer. Default impl. used if None.
            file_format_validator: File format validator. Default impl. used if None.
            file_hasher: File hasher. Default impl. used if None.
        """
        self._file_reader_writer = file_reader_writer or FileReaderWriter()
        self._file_identifier = file_identifier or FileIdentifier()
        self._role_analyzer = role_analyzer or RoleAnalyzer()
        self._file_format_validator = file_format_validator or FileFormatValidator()
        self._file_hasher = file_hasher or FileHasher()

    async def save(
        self, files: Sequence[BufferedFile], directory: Path, content_hash: str
    ) -> ProtocolSource:
        """Compute a `ProtocolSource` from buffered files and save them as files.

        The input is parsed and statically analyzed to ensure it's basically
        well-formed. For example, labware definition files must conform to the labware
        definition schema.

        Arguments:
            files: List buffered files. Do not attempt to reuse any objects
                in this list once they've been passed to the ProtocolReader.
            directory: Name of the directory to create and place files in.

        Returns:
            A ProtocolSource describing the validated protocol.

        Raises:
            ProtocolFilesInvalidError: Input file list given to the reader
                could not be validated as a protocol.
        """
        identified_files = await self._file_identifier.identify(
            files, python_parse_mode=PythonParseMode.NORMAL
        )
        role_analysis = self._role_analyzer.analyze(identified_files)
        await self._file_format_validator.validate(role_analysis.all_files)

        files_to_write = [f.original_file for f in role_analysis.all_files]
        await self._file_reader_writer.write(directory=directory, files=files_to_write)

        main_file = directory / role_analysis.main_file.original_file.name
        output_files = [
            ProtocolSourceFile(
                path=directory / f.original_file.name, role=self._map_file_role(f)
            )
            for f in role_analysis.all_files
        ]

        return ProtocolSource(
            directory=directory,
            main_file=main_file,
            content_hash=content_hash,
            files=output_files,
            config=self._map_config(role_analysis),
            robot_type=role_analysis.main_file.robot_type,
            metadata=role_analysis.main_file.metadata,
        )

    async def read_saved(
        self,
        files: Sequence[Path],
        directory: Optional[Path],
        files_are_prevalidated: bool = False,
        python_parse_mode: PythonParseMode = PythonParseMode.NORMAL,
    ) -> ProtocolSource:
        """Compute a `ProtocolSource` from protocol source files on the filesystem.

        The input is parsed and statically analyzed to ensure it's basically
        well-formed. For example, labware definition files must conform to the labware
        definition schema.

        Arguments:
            files: The files comprising the protocol.
            directory: Passed through to `ProtocolSource.directory`. Otherwise unused.
            files_are_prevalidated: Assume that the input files are valid. Skip full
                parsing and validation, doing only the minimum required to extract the
                stuff for the `ProtocolSource`. This can be 10-100x faster, but you
                should only do it with protocols that have already been validated
                by this module.
            python_parse_mode: See the documentation in `PythonParseMode`.

        Returns:
            A `ProtocolSource` describing the validated protocol.

        Raises:
            ProtocolFilesInvalidError: Input file list given to the reader
                could not be validated as a protocol.
        """
        buffered_files = await self._file_reader_writer.read(files)
        identified_files = await self._file_identifier.identify(
            files=buffered_files,
            python_parse_mode=python_parse_mode,
        )
        role_analysis = self._role_analyzer.analyze(identified_files)
        if not files_are_prevalidated:
            await self._file_format_validator.validate(role_analysis.all_files)

        # We know these paths will not be None because we supplied real Paths,
        # not AbstractInputFiles, to FileReaderWriter.
        # TODO(mc, 2022-04-01): these asserts are a bit awkward,
        # consider restructuring so they're not needed.
        assert isinstance(role_analysis.main_file.original_file.path, Path)
        assert all(
            isinstance(f.original_file.path, Path) for f in role_analysis.all_files
        )

        main_file = role_analysis.main_file.original_file.path
        content_hash = await self._file_hasher.hash(buffered_files)

        output_files = [
            ProtocolSourceFile(path=f.original_file.path, role=self._map_file_role(f))  # type: ignore[arg-type]
            for f in role_analysis.all_files
        ]

        return ProtocolSource(
            directory=directory,
            main_file=main_file,
            content_hash=content_hash,
            files=output_files,
            config=self._map_config(role_analysis),
            robot_type=role_analysis.main_file.robot_type,
            metadata=role_analysis.main_file.metadata,
        )

    @staticmethod
    def _map_file_role(file: IdentifiedFile) -> ProtocolFileRole:
        if isinstance(file, (IdentifiedJsonMain, IdentifiedPythonMain)):
            return ProtocolFileRole.MAIN
        elif isinstance(file, IdentifiedLabwareDefinition):
            return ProtocolFileRole.LABWARE
        elif isinstance(file, IdentifiedData):
            return ProtocolFileRole.DATA

    @staticmethod
    def _map_config(protocol_info: RoleAnalysis) -> ProtocolConfig:
        if isinstance(protocol_info.main_file, IdentifiedJsonMain):
            return JsonProtocolConfig(
                schema_version=protocol_info.main_file.schema_version
            )
        elif isinstance(protocol_info.main_file, IdentifiedPythonMain):
            return PythonProtocolConfig(api_version=protocol_info.main_file.api_level)
